
__all__ = [
    "center_of_mass",
    "inertia_tensors",
    "moments_of_inertia",
    "inertial_frame_derivatives",
    "translation_rotation_eigenvectors",
    "translation_rotation_projector",
    "remove_translation_rotations",
    "translation_rotation_invariant_transformation",
    "eckart_embedding",
    "rmsd_minimizing_transformation",
    "eckart_permutation",


]

import itertools, collections

import numpy as np, scipy.optimize as opt
from . import VectorOps as vec_ops
from . import PermutationOps as perm_ops
# from . import SetOps as set_ops
# from . import Misc as misc
# from . import TransformationMatrices as tf_mats

def center_of_mass(coords, masses=None):
    """Gets the center of mass for the coordinates

    :param coords:
    :type coords: CoordinateSet
    :param masses:
    :type masses:
    :return:
    :rtype:
    """

    if masses is None:
        masses = np.ones(coords.shape[-2])

    masses = masses.copy()
    masses[masses < 0] = 0

    return np.tensordot(masses / np.sum(masses), coords, axes=[0, -2])

def inertia_tensors(coords, masses=None):
    """
    Computes the moment of intertia tensors for the walkers with coordinates coords (assumes all have the same masses)

    :param coords:
    :type coords: CoordinateSet
    :param masses:
    :type masses: np.ndarray
    :return:
    :rtype:
    """

    if masses is None:
        masses = np.ones(coords.shape[-2])

    com = center_of_mass(coords, masses)
    coords = coords - com[..., np.newaxis, :]

    real_spots = masses > 0 # allow for dropping out dummy atoms
    coords = coords[..., real_spots, :]
    masses = masses[real_spots]

    d = np.zeros(coords.shape[:-1] + (3, 3), dtype=float)
    diag = vec_ops.vec_dots(coords, coords)
    d[..., (0, 1, 2), (0, 1, 2)] = diag[..., np.newaxis]
    # o = np.array([[np.outer(a, a) for a in r] for r in coords])
    o = vec_ops.vec_outer(coords, coords, axes=[-1, -1])
    tens = np.tensordot(masses, d - o, axes=[0, -3])

    return tens

def inertial_frame_derivatives(coords, masses=None, sel=None):
    if masses is None:
        masses = np.ones(coords.shape[-2])

    masses = np.asanyarray(masses)
    coords = np.asanyarray(coords)
    real_pos = masses > 0
    if sel is not None:
        real_pos = np.intersect1d(sel, real_pos)

    smol = coords.ndim == 2
    if smol:
        coords = coords[np.newaxis]
    base_shape = coords.shape[:-2]
    coords = coords.reshape((-1,) + coords.shape[-2:])
    if masses.ndim == 1:
        masses = masses[np.newaxis]
    else:
        masses = np.reshape(masses, (-1, masses.shape[-1]))

    masses = masses[..., real_pos]
    coords = coords[..., real_pos, :]

    masses = np.sqrt(masses)
    carts = masses[..., :, np.newaxis] * coords  # mass-weighted Cartesian coordinates

    ### compute basic inertia tensor derivatives
    # first derivs are computed as a full (nAt, 3, I_rows (3), I_cols (3)) tensor
    # and then reshaped to (nAt * 3, I_rows, I_cols)
    eyeXeye = np.eye(9).reshape(3, 3, 3, 3).transpose((2, 0, 1, 3))
    I0Y_1 = np.tensordot(carts, eyeXeye, axes=[2, 0])

    nAt = carts.shape[1]
    nY = nAt * 3
    I0Y_21 = (
            np.reshape(np.eye(3), (9,))[np.newaxis, :, np.newaxis]
            * carts[:, :, np.newaxis, :]
    )  # a flavor of outer product
    I0Y_21 = I0Y_21.reshape((-1, nAt, 3, 3, 3))
    I0Y_2 = (I0Y_21 + I0Y_21.transpose((0, 1, 2, 4, 3)))
    I0Y = 2 * I0Y_1 - I0Y_2
    I0Y = I0Y.reshape(base_shape + (nY, 3, 3))

    # second derivatives are 100% independent of coorinates
    # only the diagonal blocks are non-zero, so we compute that block
    # and then tile appropriately
    keyXey = np.eye(9).reshape(3, 3, 3, 3)
    I0YY_nn = 2 * eyeXeye - (keyXey + keyXey.transpose((0, 1, 3, 2)))
    I0YY = np.zeros((nAt, 3, nAt, 3, 3, 3))
    for n in range(nAt):
        I0YY[n, :, n, :, :, :] = I0YY_nn
    I0YY = I0YY.reshape((nY, nY, 3, 3))
    I0YY = np.broadcast_to(I0YY[np.newaxis, :, :, :, :], (carts.shape[0],) + I0YY.shape)
    I0YY = np.reshape(I0YY, base_shape + (nY, nY, 3, 3))

    if smol:
        I0Y = I0Y[0]
        I0YY = I0YY[0]

    return [I0Y, I0YY]

def moments_of_inertia(coords, masses=None):
    """
    Computes the moment of inertia tensor for the walkers with coordinates coords (assumes all have the same masses)

    :param coords:
    :type coords: CoordinateSet
    :param masses:
    :type masses: np.ndarray
    :return:
    :rtype:
    """

    if coords.ndim == 1:
        raise ValueError("can't get moment of inertia for single point (?)")
    elif coords.ndim == 2:
        multiconfig = False
        coords = coords[np.newaxis]
        extra_shape = None
    else:
        multiconfig = True
        extra_shape = coords.shape[:-2]
        coords = coords.reshape((np.prod(extra_shape),) + coords.shape[-2:])


    if masses is None:
        masses = np.ones(coords.shape[-2])

    massy_doop = inertia_tensors(coords, masses)
    moms, axes = np.linalg.eigh(massy_doop)
    # a = axes[..., :, 0]
    # c = axes[..., :, 2]
    # b = nput.vec_crosses(a, c)  # force right-handedness to avoid inversions
    # axes[..., :, 1] = b
    a = axes[..., :, 0]
    b = axes[..., :, 1]
    c = vec_ops.vec_crosses(b, a)  # force right-handedness to avoid inversions
    axes[..., :, 2] = c
    dets = np.linalg.det(axes) # ensure we have true rotation matrices to avoid inversions
    axes[..., :, 2] /= dets[..., np.newaxis]

    if multiconfig:
        moms = moms.reshape(extra_shape + (3,))
        axes = axes.reshape(extra_shape + (3, 3))
    else:
        moms = moms[0]
        axes = axes[0]
    return moms, axes

def translation_rotation_eigenvectors(coords, masses=None, mass_weighted=True):
    """
    Returns the eigenvectors corresponding to translations and rotations
    in the system

    :param coords:
    :type coords:
    :param masses:
    :type masses:
    :return:
    :rtype:
    """


    if masses is None:
        masses = np.ones(coords.shape[-2])
    masses = np.asanyarray(masses)

    n = len(masses)
    # explicitly put masses in m_e from AMU
    # masses = UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass") * masses
    mT = np.sqrt(np.sum(masses))
    mvec = np.sqrt(masses)

    # no_derivs = order is None
    # if no_derivs:
    #     order = 0

    smol = coords.ndim == 2
    if smol:
        coords = coords[np.newaxis]
    # base_shape = None
    # if coords.ndim > 3:
    base_shape = coords.shape[:-2]
    coords = coords.reshape((-1,) + coords.shape[-2:])

    M = np.kron(mvec / mT, np.eye(3)).T  # translation eigenvectors
    mom_rot, ax_rot = moments_of_inertia(coords, masses)
    # if order > 0:
    #     base_tensor = StructuralProperties.get_prop_inertia_tensors(coords, masses)
    #     mom_expansion = StructuralProperties.get_prop_inertial_frame_derivatives(coords, masses)
    #     inertia_expansion = [base_tensor] + mom_expansion
    #     sqrt_expansion = nput.matsqrt_deriv(inertia_expansion, order)
    #     inv_rot_expansion = nput.matinv_deriv(sqrt_expansion, order=order)
    #     inv_rot_2 = inv_rot_expansion[0]
    # else:
    good_ax = np.abs(mom_rot) > 1e-8
    nonlinear = good_ax.flatten().all()
    if nonlinear:
        inv_mom_2 = vec_ops.vec_tensordiag(1 / np.sqrt(mom_rot))
        inv_rot_2 = vec_ops.vec_tensordot(
            ax_rot,
            vec_ops.vec_tensordot(
                ax_rot,
                inv_mom_2,
                shared=1,
                axes=[-1, -1]
            ),
            shared=1,
            axes=[-1, -1]
        )
    else:
        all_linear = len(good_ax)==0 or np.sum(np.abs(np.diff(good_ax, axis=0))) == 0
        if all_linear:
            good_ax = good_ax[0]
            inv_mom_2 = vec_ops.vec_tensordiag(1 / np.sqrt(mom_rot[:, good_ax,]))
            inv_rot_2 = vec_ops.vec_tensordot(
                ax_rot[:, :, good_ax],
                vec_ops.vec_tensordot(
                    ax_rot[:, :, good_ax],
                    inv_mom_2,
                    shared=1,
                    axes=[-1, -1]
                ),
                shared=1,
                axes=[-1, -1]
            )
        else:
            inv_rot_2 = []
            for sel,moms,rot in zip(good_ax, mom_rot, ax_rot):
                inv_mom_2 = vec_ops.vec_tensordiag(1 / np.sqrt(moms[sel,]))
                subinv = rot[:, sel] @ inv_mom_2 @ rot[:, sel].T
                rem = np.array([0, 1, 2])[np.logical_not(sel)]
                subinv[:, rem] = rot.T[:, rem]
                inv_rot_2.append(subinv)
            inv_rot_2 = np.array(inv_rot_2)

    com = center_of_mass(coords, masses)
    com = np.expand_dims(com, 1) # ???
    shift_crds = mvec[np.newaxis, :, np.newaxis] * (coords - com[: np.newaxis, :])
    cos_rot = perm_ops.levi_cevita_dot(3, inv_rot_2, axes=[0, -1], shared=1) # kx3bx3cx3j
    R = vec_ops.vec_tensordot(
        shift_crds, cos_rot,
        shared=1,
        axes=[-1, 1]
    ).reshape((coords.shape[0], 3 * n, 3))  # rotations

    if nonlinear:
        freqs = np.concatenate([
            np.broadcast_to([[1e-14, 1e-14, 1e-14]], mom_rot.shape),
            (1 / (2 * mom_rot))
            # this isn't right, I'm totally aware, but I think the frequency is supposed to be zero anyway and this
            # will be tiny
        ], axis=-1)
        M = np.broadcast_to(M[np.newaxis], R.shape)
        eigs = np.concatenate([M, R], axis=2)
    elif all_linear:
        mom_rot = mom_rot[:, good_ax]
        freqs = np.concatenate([
            np.broadcast_to([[1e-14, 1e-14, 1e-14]], R.shape[:1] + (3,)),
            (1 / (2 * mom_rot))
            # this isn't right, I'm totally aware, but I think the frequency is supposed to be zero anyway and this
            # will be tiny
        ], axis=-1)
        R = R[:, :, good_ax]
        M = np.broadcast_to(M[np.newaxis], R.shape[:1] + M.shape)
        eigs = np.concatenate([M, R], axis=2)
    else:
        freqs = []
        eigs = []
        for sel, moms, rot in zip(good_ax, mom_rot, R):
            f = np.concatenate([[1e-14, 1e-14, 1e-14], 1/(2*mom_rot[sel,])])
            r = rot[:, sel]
            freqs.append(f)
            eigs.append(np.concatenate([M, r], axis=1))

    if not mass_weighted:
        W = np.diag(np.repeat(1/np.sqrt(masses), 3))
        eigs = np.moveaxis(np.tensordot(eigs, W, axes=[-2, 0]), -1, -2)

    if smol:
        eigs = eigs[0]
        freqs = freqs[0]
    else:
        eigs = eigs.reshape(base_shape + eigs.shape[1:])
        freqs = freqs.reshape(base_shape + freqs.shape[1:])

    return freqs, eigs

def translation_rotation_projector(coords, masses=None, mass_weighted=False, return_modes=False,
                                   orthonormal=True
                                   ):
    if masses is None:
        masses = np.ones(coords.shape[-2])
    _, tr_modes = translation_rotation_eigenvectors(coords, masses, mass_weighted=mass_weighted)
    if orthonormal:
        if not mass_weighted:
            tr_modes = vec_ops.vec_normalize(tr_modes, axis=0)
        projector = vec_ops.orthogonal_projection_matrix(tr_modes, orthonormal=True)
    else:
        shared = tr_modes.ndim - 2
        eye = vec_ops.identity_tensors(tr_modes.shape[:-2], tr_modes.shape[-2])
        projector = eye - vec_ops.vec_tensordot(tr_modes, tr_modes, axes=[-1, -1], shared=shared)

    if return_modes:
        return projector, tr_modes
    else:
        return projector


def remove_translation_rotations(expansion, coords, masses=None, mass_weighted=False):
    projector = translation_rotation_projector(coords, masses=masses, mass_weighted=mass_weighted)
    shared = projector.ndim - 2

    proj_expansion = []
    for n,d in enumerate(expansion):
        for ax in range(n+1):
            d = vec_ops.vec_tensordot(projector, d, axes=[-1, shared+ax], shared=shared)
        proj_expansion.append(d)

    return proj_expansion

def translation_rotation_invariant_transformation(
        coords, masses=None,
        mass_weighted=True,
        strip_embedding=True
):

    if masses is None:
        masses = np.ones(coords.shape[-2])

    A, L_tr = translation_rotation_projector(coords, masses, mass_weighted=True, return_modes=True)
    base_shape = A.shape[:-2]
    A = A.reshape((-1,) + A.shape[-2:])
    L_tr = L_tr.reshape((-1,) + L_tr.shape[-2:])
    evals, tf = np.linalg.eigh(A)
    zero_pos = np.abs(evals) < 1e-4# the rest should be 1
    zero_counts = np.sum(zero_pos, axis=1)
    ucounts = np.min(zero_counts)

    tf[:, :, :ucounts] = L_tr
    if strip_embedding:
        # nzpos = np.where(np.abs(evals) > 1e-4) # the rest should be 1
        # nzpos = nzpos[:-1] + (slice(None),) + nzpos[-1:]
        tf = tf[:, :, ucounts:]
    else:
        tf[:, :, :ucounts] = L_tr

    inv = np.moveaxis(tf, -1, -2)
    if not mass_weighted:
        W = np.diag(np.repeat(np.sqrt(masses), 3))
        tf = np.moveaxis(
            np.tensordot(W, tf, axes=[0, 1]),
            1, 0
        )

        W = np.diag(np.repeat(1/np.sqrt(masses), 3))
        inv = np.tensordot(inv, W, axes=[2, 0])

    tf = tf.reshape(base_shape + tf.shape[-2:])
    inv = inv.reshape(base_shape + inv.shape[-2:])

    return tf, inv

EmbeddingData = collections.namedtuple("PrincipleAxisData", ['coords', 'com', 'axes'])
EckartData = collections.namedtuple('EckartData', ['rotations', 'coordinates', 'reference_data', 'coord_data'])
def principle_axis_embedded_coords(coords, masses=None):
    """
    Returns coordinate embedded in the principle axis frame

    :param coords:
    :type coords:
    :param masses:
    :type masses:
    :return:
    :rtype:
    """
    com = center_of_mass(coords, masses)
    # crds_ = coords
    coords = coords - com[..., np.newaxis, :]
    moms, pax_axes = moments_of_inertia(coords, masses)
    # pax_axes = np.swapaxes(pax_axes, -2, -1)
    coords = np.matmul(coords, pax_axes)

    return EmbeddingData(coords, com, pax_axes)

def _prep_eckart_data(ref, coords, masses, in_paf=False, sel=None):

    if masses is None:
        masses = np.ones(coords.shape[-2])

    if coords.ndim == 2:
        coords = np.broadcast_to(coords, (1,) + coords.shape)

    if not in_paf:
        coords, com, pax_axes = principle_axis_embedded_coords(coords, masses)
        ref, ref_com, ref_axes = principle_axis_embedded_coords(ref, masses)
        # raise ValueError(ref)
    else:
        com = pax_axes = None
        ref_com = ref_axes = None

    og_coords = coords
    if sel is not None:
        coords = coords[..., sel, :]
        masses = masses[sel]
        ref = ref[..., sel, :]

    real_pos = masses > 0
    # print(real_pos)
    # og_coords = coords
    coords = coords[..., real_pos, :]
    masses = masses[real_pos,]
    og_ref = ref
    ref = ref[..., real_pos, :]

    if ref.ndim == 2:
        ref = ref[np.newaxis]
    if ref.shape[0] > 1 and ref.shape[0] < coords.shape[0]:  # TODO: make less hacky
        # need to make them broadcast together and we assume
        # we have an extra stack of coords
        n_sys = coords.shape[0]
        ref = np.reshape(
            np.broadcast_to(
                ref[np.newaxis],
                (n_sys // ref.shape[0],) + ref.shape
            ),
            (n_sys,) + ref.shape[1:]
        )
        ref_axes = np.reshape(
            np.broadcast_to(
                ref_axes[np.newaxis],
                (n_sys // ref_axes.shape[0],) + ref_axes.shape
            ),
            (n_sys,) + ref_axes.shape[1:]
        )
        ref_com = np.reshape(
            np.broadcast_to(
                ref_com[np.newaxis],
                (n_sys // ref_com.shape[0],) + ref_com.shape
            ),
            (n_sys,) + ref_com.shape[1:]
        )

    return (ref, ref_com, ref_axes), (coords, com, pax_axes), masses, (og_ref, og_coords)

def _eckart_embedding(ref, coords,
                      masses=None,
                      sel=None,
                      in_paf=False,
                      planar_ref_tolerance=1e-6,
                      proper_rotation=False,
                      permutable_groups=None,
                      transform_coordinates=True
                      ):
    """
    Generates the Eckart rotation that will align ref and coords, assuming initially that `ref` and `coords` are
    in the principle axis frame

    :param masses:
    :type masses:
    :param ref:
    :type ref:
    :param coords:
    :type coords: np.ndarray
    :return:
    :rtype:
    """

    coords = np.asanyarray(coords)
    og_og_coords = coords
    ref = np.asanyarray(ref)
    og_og_ref = ref

    (ref, ref_com, ref_axes), (coords, com, pax_axes), masses, (og_ref, og_coords) = _prep_eckart_data(
        ref, coords, masses,
        sel=sel, in_paf=in_paf
    )

    # needs to be updated for the multiple reference case?
    # TODO: make sure that we broadcast this correctly to check if all or
    #       none of the reference structures are planar
    planar_ref = np.allclose(ref[0][:, 2], 0., atol=planar_ref_tolerance)

    if permutable_groups is not None:
        if sel is not None:
            reindexing = {a:i for i,a in enumerate(sel)}
            permutable_groups = [
                [reindexing[a] for a in g if a in reindexing]
                for g in permutable_groups
            ]
            permutable_groups = [g for g in permutable_groups if len(g) > 1]
        rem_atoms = np.delete(np.arange(len(masses)), np.concatenate(permutable_groups))
        permutable_groups = list(permutable_groups) + rem_atoms[:, np.newaxis].tolist()
    if not planar_ref:
        # generate pair-wise product matrix
        if permutable_groups is None:
            A = np.tensordot(
                masses / np.sum(masses),
                ref[:, :, :, np.newaxis] * coords[:, :, np.newaxis, :],
                axes=[0, 1]
            )
        else:
            mw_scaling = np.expand_dims(np.sqrt(masses) / np.sum(masses), [0, 2])
            mw_ref = ref * mw_scaling
            mw_coords = coords * mw_scaling
            A = sum(
                np.sum(
                    mw_ref[:, g, :, np.newaxis],
                    mw_coords[:, g, np.newaxis, :],
                    axis=1
                ) for g in permutable_groups
            )
        # take SVD of this
        U, S, V = np.linalg.svd(A)
        rot = np.matmul(U, V)
        if proper_rotation:
            dets = np.linalg.det(rot)
            inversions = np.where(dets < 0)
            V[inversions, :, 2] *= -1
            rot = np.matmul(U, V)

    else:
        # generate pair-wise product matrix but only in 2D
        if permutable_groups is None:
            F = ref[:, :, :2, np.newaxis] * coords[:, :, np.newaxis, :2]
            A = np.tensordot(masses / np.sum(masses), F, axes=[0, 1])
        else:
            mw_scaling = np.expand_dims(np.sqrt(masses) / np.sum(masses), [0, 2])
            mw_ref = ref * mw_scaling
            mw_coords = coords * mw_scaling
            A = sum(
                np.sum(
                    mw_ref[:, g, :2, np.newaxis],
                    mw_coords[:, g, np.newaxis, :2],
                    axis=1
                ) for g in permutable_groups
            )

        U, S, V = np.linalg.svd(A)
        base_rot = np.matmul(U, V)
        if proper_rotation:
            dets = np.linalg.det(base_rot)
            inversions = np.where(dets < 0)
            V[inversions, :, 1] *= -1
            base_rot = np.matmul(U, V)

        rot = np.broadcast_to(np.eye(3, dtype=float), (len(coords), 3, 3)).copy()
        rot[..., :2, :2] = base_rot

    if transform_coordinates:
        # crd is in _its_ principle axis frame, so now we transform it using ek_rot
        ek_rot = np.swapaxes(rot, -2, -1)
        coords = og_coords @ ek_rot
        # now we rotate this back to the reference frame
        coords = coords @ np.swapaxes(ref_axes if ref_axes.ndim > 2 else ref_axes[np.newaxis], -2, -1)
        # and then shift so the COM doesn't change
        coords = coords + (ref_com if ref_com.ndim > 1 else ref_com[np.newaxis, np.newaxis, :])
    else:
        coords = None

    base_coord_shape = og_og_coords.shape[:-2]
    coords = coords.reshape(base_coord_shape + coords.shape[-2:])
    og_coords = og_coords.reshape(base_coord_shape + og_coords.shape[-2:])
    com = com.reshape(base_coord_shape + com.shape[-1:])
    pax_axes = pax_axes.reshape(base_coord_shape + pax_axes.shape[-2:])
    rot = rot.reshape(base_coord_shape + rot.shape[-2:])
    base_ref_shape = og_og_ref.shape[:-2]
    og_ref = og_ref.reshape(base_ref_shape + og_ref.shape[-2:])
    ref_com = ref_com.reshape(base_ref_shape + ref_com.shape[-1:])
    ref_axes = ref_axes.reshape(base_ref_shape + ref_axes.shape[-2:])
    return EckartData(
        rot,
        coords,
        EmbeddingData(og_ref, ref_com, ref_axes),
        EmbeddingData(og_coords, com, pax_axes)
    )

def eckart_embedding(ref, coords,
                     masses=None,
                     sel=None,
                     in_paf=False,
                     planar_ref_tolerance=1e-6,
                     proper_rotation=False,
                     permutable_groups=None,
                     transform_coordinates=True):
    # if permutable_groups is None:
    return _eckart_embedding(
        ref, coords,
        masses=masses,
        sel=sel,
        in_paf=in_paf,
        planar_ref_tolerance=planar_ref_tolerance,
        proper_rotation=proper_rotation,
        permutable_groups=permutable_groups,
        transform_coordinates=transform_coordinates
    )
rmsd_minimizing_transformation = eckart_embedding

def eckart_permutation(
        ref, coords,
        masses=None,
        sel=None,
        in_paf=False,
        prealign=False,
        planar_ref_tolerance=1e-6,
        proper_rotation=False,
        permutable_groups=None
):

    ref = np.asanyarray(ref)
    og_og_ref = ref
    coords = np.asanyarray(coords)
    og_og_coords = coords
    if prealign:
        embedding_data = eckart_embedding(ref, coords,
                                          masses=masses,
                                          sel=sel,
                                          in_paf=in_paf,
                                          planar_ref_tolerance=planar_ref_tolerance,
                                          proper_rotation=proper_rotation,
                                          permutable_groups=permutable_groups)
        # ref = embedding_data.reference_data.coords
        coords = embedding_data.coordinates

    (ref, ref_com, ref_axes), (coords, com, pax_axes), masses, (og_ref, og_coords) = _prep_eckart_data(
        ref, coords, masses,
        sel=sel, in_paf=True
    )

    mw_scaling = np.expand_dims(np.sqrt(masses) / np.sum(masses), [0, 2])
    mw_ref = ref * mw_scaling
    # mw_coords = coords * mw_scaling

    if permutable_groups is not None:
        if sel is not None:
            reindexing = {a: i for i, a in enumerate(sel)}
            permutable_groups = [
                [reindexing[a] for a in g if a in reindexing]
                for g in permutable_groups
            ]
            permutable_groups = [g for g in permutable_groups if len(g) > 1]
        rem_atoms = np.delete(np.arange(len(masses)), np.concatenate(permutable_groups))
        permutable_groups = list(permutable_groups) + rem_atoms[:, np.newaxis].tolist()
    else:
        permutable_groups = [np.arange(mw_ref.shape[-2])]

    base_perm = np.repeat(np.arange(og_ref.shape[-2])[np.newaxis], ref.shape[0], axis=0)
    if sel is not None:
        perm = base_perm[:, sel]
    else:
        perm = base_perm
    base_shape = ref.shape[:-2]
    for p in permutable_groups:
        sub_data = eckart_embedding(ref, coords,
                                    masses=masses,
                                    sel=p,
                                    in_paf=in_paf,
                                    planar_ref_tolerance=planar_ref_tolerance,
                                    proper_rotation=proper_rotation,
                                    permutable_groups=[p])
        mw_coords = sub_data.coordinates * mw_scaling

        p = np.asanyarray(p)
        n = len(p)
        dists = np.zeros(base_shape + (n, n))
        rows, cols = np.triu_indices(n, k=0)
        dist_triu = np.linalg.norm(
            mw_coords[..., rows, :] - mw_ref[..., p[cols], :],
            axis=-1
        )
        dists[..., rows, cols] = dist_triu
        dists[..., cols, rows] = dist_triu

        for k,cost in enumerate(dists):
            _, sub_perm = opt.linear_sum_assignment(cost)
            perm[k][p] = p[sub_perm]
    if sel is not None:
        base_perm[:, sel] = perm

    targ_shape = og_og_coords.shape[:-2]
    return base_perm.reshape(targ_shape + (-1,))


def eckart_displacement_coords(coords, ref, masses=None, **embedding_parameters):
    ref_embedding, _ = translation_rotation_invariant_transformation(ref,
                                                                  masses=masses,
                                                                  mass_weighted=False
                                                                  )
    base_embedding = eckart_embedding(ref, coords, masses=masses, **embedding_parameters)
    return np.reshape(
        ref_embedding @ base_embedding.coordinates[..., np.newaxis],
        base_embedding.coordinates.shape
    )




import numpy as np
import itertools
import collections
from .. import Numputils as nput
from ..Graphs import EdgeGraph

__all__ = [
    "get_stretch_angles",
    "get_angle_dihedrals",
    "get_angle_stretches",
    "get_dihedral_stretches",
    "get_stretch_angle_dihedrals",
    "get_stretch_coordinate_system",
    "PrimitiveCoordinatePicker"
]


def get_stretch_angles(stretches):
    angles = []
    for i,(sa,sb) in enumerate(stretches):
        for sc,sd in stretches[i+1:]:
            if sa == sc:
                if sb == sd: continue
                angles.append((sb, sa, sd))
            elif sa == sd:
                if sb == sc: continue
                angles.append((sb, sa, sc))
            elif sb == sc:
                angles.append((sa, sb, sd))
            elif sb == sd:
                angles.append((sa, sb, sc))
    return angles
def get_stretch_angle_dihedrals(stretches, angles):
    dihedrals = []
    for sa,sb in stretches:
        for ba,bc,bd in angles:
            if sa in (ba,bc,bd) and sb in (ba,bc,bd): continue
            # enumerate for simplicity & avoiding try/except
            if sa == ba:
                dihedrals.append(
                    (sb, ba, bc, bd)
                )
            elif sa == bc:
                dihedrals.append(
                    (ba, sb, bc, bd)
                )
            elif sa == bd:
                dihedrals.append(
                    (ba, bc, bd, sb)
                )
            elif sb == ba:
                dihedrals.append(
                    (sa, ba, bc, bd)
                )
            elif sb == bc:
                dihedrals.append(
                    (ba, sa, bc, bd)
                )
            elif sb == bd:
                dihedrals.append(
                    (ba, bc, bd, sa)
                )
    return dihedrals
def get_angle_stretches(angles):
    return [
        s
        for a,b,c in angles
        for s in [(a, b), (b, c)]
    ]
def get_dihedral_stretches(dihedrals):
    return [
        s
        for a,b,c,d in dihedrals
        for s in [(a, b), (b, c), (c,d)]
    ]
def get_angle_dihedrals(angles):
    dihedrals = []
    for i, (aa, ab, ac) in enumerate(angles):
        for ad, ae, af in angles[i + 1:]:
            if ae == ac:
                if ad == ab:
                    if af == aa: continue
                    dihedrals.append((aa, ab, ac, af))
                elif af == ab:
                    if ad == aa: continue
                    dihedrals.append((aa, ab, ac, ad))
            elif ae == aa:
                if ad == ab:
                    if af == ac: continue
                    dihedrals.append((af, aa, ab, ac))
                elif af == ab:
                    if ad == ac: continue
                    dihedrals.append((ad, aa, ab, ac))
    return dihedrals

def get_stretch_coordinate_system(stretches):
    angles = get_stretch_angles(stretches)
    dihedrals = get_angle_dihedrals(angles)
    return stretches,angles,dihedrals


class PrimitiveCoordinatePicker:

    light_atom_types = {"H", "D"}
    def __init__(self, atoms, bonds, base_coords=None, rings=None, fragments=None, light_atoms=None, backbone=None, neighbor_count=3):
        self.graph = EdgeGraph(atoms, bonds)
        if rings is None:
            rings = self.graph.get_rings()
        self.rings = rings
        self.ring_atoms = {k:n for n,rats in enumerate(rings) for k in rats}
        if fragments is None:
            fragments = self.graph.get_fragments()
        self.fragments = fragments
        self.backbone = backbone
        self.light_atoms = (
            [
                i for i, l in enumerate(self.graph.labels)
                if l in self.light_atom_types
            ]
                if light_atoms is None else
            light_atoms
        )
        self.base_coords = list(base_coords) if base_coords is not None else []
        self.neighbors = neighbor_count
        self._coords = None
    @property
    def coords(self):
        if self._coords is None:
            self._coords = tuple(self.generate_coords())
        return self._coords
    def generate_coords(self):
        coords = []
        for ring in self.rings:
            coords.extend(self.ring_coordinates(ring))
        for i,r1 in enumerate(self.rings):
            for r2 in self.rings[i+1:]:
                coords.extend(self.fused_ring_coordinates(r1, r2))
        for i,r1 in enumerate(self.fragments):
            for r2 in self.fragments[i+1:]:
                coords.extend(self.fragment_connection_coords(r1, r2))
        for a in range(len(self.graph.labels)):
            if a not in self.ring_atoms:
                symm_c = self.symmetry_coords(a, neighborhood=self.neighbors, backbone=self.backbone)
                coords.extend(symm_c)
        coords = self.base_coords + coords
        return self.prune_excess_coords(coords)

    @classmethod
    def canonicalize_coord(cls, coord):
        dupes = len(np.unique(coord)) < len(coord)
        if dupes: return None
        if len(coord) == 2:
            i,j = coord
            if i > j:
                coord = (j, i)
        elif len(coord) == 3:
            i,j,k = coord
            if i > k:
                coord = (k, j, i)
        elif len(coord) == 4:
            i, j, k, l = coord
            if i > l:
                coord = (l, k, j, i)
        elif coord[0] > coord[-1]:
            coord = tuple(reversed(coord))
        return coord

    @classmethod
    def prep_unique_coords(cls, coords):
        _coords = []
        _cache = set()
        for x in coords:
            x = cls.canonicalize_coord(x)
            if x not in _cache:
                _coords.append(x)
                _cache.add(x)
        return coords

    @classmethod
    def prune_excess_coords(cls, coord_set, canonicalized=False):
        if not canonicalized:
            coord_set = [cls.canonicalize_coord(c) for c in coord_set]
            coord_set = [c for c in coord_set if c is not None]
        dupe_set = set()
        coords = []
        coord_counts = {}
        for coord in coord_set:
            if coord in dupe_set: continue
            dupe_set.add(coord)
            if len(coord) == 4:
                i,j,k,l = coord
                if (i, k, j, l) in dupe_set: continue
                # only one choice of ordering for shared bond, even though implies different exterior bonds
            elif len(coord) == 3:
                i,j,k = coord
                if ( # can't have all three angles, dealing with the ordering manually
                    (k < j and (i, k, j) in dupe_set and (k, i, j) in dupe_set)
                    or (k > j and (j, i, k) in dupe_set) and ((i, k, j) in dupe_set or (j, k, i) in dupe_set)
                ): continue
            coords.append(coord)
            # key = tuple(sorted(coord))
            # coord_counts[key] = coord_counts.get(key, 0) + 1
            # if len(coord_counts)

        return coords

    @classmethod
    def ring_coordinates(cls, ring_atoms):
        # ordered as pair-wise bonded

        bonds = list(zip(
            ring_atoms, ring_atoms[1:] + ring_atoms[:1]
        ))
        angles = list(zip(
            ring_atoms, ring_atoms[1:] + ring_atoms[:1], ring_atoms[2:] + ring_atoms[:2]
        ))
        dihedrals = list(zip(
            ring_atoms, ring_atoms[1:] + ring_atoms[:1],
                        ring_atoms[2:] + ring_atoms[:2],
                        ring_atoms[3:] + ring_atoms[:3],
        ))

        return bonds + angles + dihedrals

    fused_ring_dispatch_table = {

    }
    @classmethod
    def _fused_dispatch(cls):
        return dict({
            0:cls.unfused_ring_coordinates,
            1:cls.pivot_fused_ring_coordinates,
            2:cls.simple_fused_ring_coordinates
        }, **cls.fused_ring_dispatch_table)
    @classmethod
    def unfused_ring_coordinates(cls, ring_atoms1, ring_atoms2, shared_atoms, shared_indices1, shared_indices2):
        return []
    @classmethod
    def pivot_fused_ring_coordinates(cls, ring_atoms1, ring_atoms2, shared_atoms, shared_indices1, shared_indices2):
        p = shared_atoms[0]
        i = shared_indices1[0]
        n = len(ring_atoms1)
        j = shared_indices2[0]
        m = len(ring_atoms2)

        # add in all relative angles
        ip1 = (i+1) % n
        jp1 = (j+1) % m
        return [
            (ring_atoms1[i-1], p, ring_atoms2[j-1]),
            (ring_atoms1[ip1], p, ring_atoms2[j-1]),
            (ring_atoms1[i-1], p, ring_atoms2[jp1]),
            (ring_atoms1[ip1], p, ring_atoms2[jp1])
        ]

    @classmethod
    def simple_fused_ring_coordinates(cls, ring_atoms1, ring_atoms2, shared_atoms, shared_indices1, shared_indices2):
        j, k = shared_atoms
        j1, k1 = shared_indices1
        j2, k2 = shared_indices2
        # we want relative orientation indices for both rings
        inds = []
        if j1 > k1:
            j1, k1 = k1, j1
        if j2 > k2:
            j2, k2 = k2, j2
        i1 = ring_atoms1[j1 - 1]; l1 = ring_atoms1[(k1 + 1) % len(ring_atoms1)]
        i2 = ring_atoms2[j2 - 1]; l2 = ring_atoms2[(k2 + 1) % len(ring_atoms2)]

        return [
            (i1, j, k, l2),
            (i2, j, k, l1)
        ]

    @classmethod
    def fused_ring_coordinates(cls, ring_atoms1, ring_atoms2):
        shared_atoms, _, _, r1_indices, r2_indices = nput.intersection(ring_atoms1, ring_atoms2, return_indices=True)
        if len(shared_atoms) == 0:
            return []
        else:
            n = len(shared_atoms)
            coord_func = cls._fused_dispatch().get(n)
            if coord_func is None:
                raise ValueError(f"can't deal with fused rings with {n} shared atoms")
            return coord_func(ring_atoms1, ring_atoms2, shared_atoms, r1_indices, r2_indices)

    def fragment_connection_coords(self, frag_1, frag_2):
        heavy_frag1 = set(frag_1) - set(self.light_atoms)
        if len(heavy_frag1) > 2:
            frag_1 = list(sorted(heavy_frag1))
        heavy_frag2 = set(frag_2) - set(self.light_atoms)
        if len(heavy_frag2) > 2:
            frag_2 = list(sorted(heavy_frag2))

        coords = []
        coords.append((frag_1[0], frag_2[0]))
        if len(frag_2) > 1:
            coords.append((frag_1[0], frag_2[0], frag_2[1]))
        if len(frag_1) > 1:
            coords.append((frag_1[1], frag_1[0], frag_2[0]))
        if len(frag_2) > 1 and len(frag_1) > 1:
            coords.append((frag_1[1], frag_1[0], frag_2[0], frag_2[1]))
        if len(frag_1) > 2:
            coords.append((frag_1[2], frag_1[1], frag_1[0], frag_2[0]))
        if len(frag_2) > 2:
            coords.append((frag_1[0], frag_2[0], frag_2[1], frag_2[2]))

        return coords

    def get_neighborhood_symmetries(self, atoms, ignored=None, neighborhood=3):
        graphs = [
            self.graph.neighbor_graph(a, ignored=ignored, num=neighborhood)
            for a in atoms
        ]
        rows, cols = np.triu_indices(len(graphs), k=1)
        return [graphs[r] == graphs[c] for r,c in zip(rows, cols)]

    def chain_coords(self, R, y):
        coords = []
        if len(R) > 0:
            coords.append((y, R[-1]))
        if len(R) > 1:
            coords.append((y, R[-1], R[-2]))
        if len(R) > 2:
            coords.append((y, R[-1], R[-2], R[-3]))
        return coords

    def RYX2_coords(self, R, y, X):
        coords = []
        coords.extend((y, x) for x in X)
        coords.extend(
            (X[i], y, X[j])
            for i, j in itertools.combinations(range(3), 2)
        )
        if len(R) > 0:
            coords.append((R[-1], y))
            # add in RYX angles
            coords.extend(
                (R[-1], y, x)
                for x in X
            )
        if len(R) > 1:
            # add in RRY angle
            coords.append(
                (R[-2], R[-1], y)
            )
        if len(R) > 2:
            # add in RRRY dihedral
            coords.append(
                (R[-3], R[-2], R[-1], y)
            )

        return coords

    def RYX3_coords(self, R, y, X):
        coords = []
        coords.extend((y, x) for x in X)
        coords.extend(
            (X[i], y, X[j])
            for i, j in itertools.combinations(range(3), 2)
        )
        if len(R) > 0:
            coords.append((R[-1], y))
            # add in RYX angles
            coords.extend(
                (R[-1], y, x) for x in X
            )
        if len(R) > 1:
            # add in RRY angle
            coords.append(
                (R[-2], R[-1], y)
            )
        if len(R) > 2:
            # add in RRRY dihedral
            coords.append(
                (R[-3], R[-2], R[-1], y)
            )

        return coords

    def get_precedent_chains(self, atom, num_precs=2, ring_atoms=None, light_atoms=None, ignored=None, backbone=None):
        chains = []
        visited = set([] if ignored is None else ignored)
        ring_atoms = set(self.ring_atoms if ring_atoms is None else ring_atoms)
        light_atoms = set(self.light_atoms if light_atoms is None else light_atoms)
        if backbone is not None:
            backbone = set(backbone)

        visited = visited #| ring_atoms # | light_atoms

        # do a dfs exploration up to the given depth over non-ring, heavy atoms
        queue = collections.deque([[[], 0, atom]])
        while queue:
            chain, depth, root = queue.pop()
            neighbors = self.graph.map[root]
            visited.add(root)
            branches = neighbors - visited
            if len(branches) == 0:
                chains.append(chain)
            elif depth == num_precs - 1:
                chains.extend(chain + [n] for n in branches)
            else:
                queue.extend([chain + [n], depth+1, n] for n in branches)

            #
            # queue.extend(rem)
            # while len(rem) == 0 and queue:
            #     # walk up dfs tree until we find a branch with nodes that work
            #     root = queue.pop()
            #     if root not in visited:
            #         rem = {root}
            # else:
            #     if len(rem) == 0: rem = neighbors - visited - light_atoms
            #     # if len(rem) == 0: rem = neighbors - visited
            #     if len(rem) == 0: break
            #
            #     if backbone is not None:
            #         bb_chain = rem & backbone
            #     else:
            #         bb_chain = []
            #     if len(bb_chain) > 0:
            #         atom = min(bb_chain)
            #     else:
            #         atom = min(rem)
            #     chain.append(atom)

        # chain = list(reversed(chain))

        return [list(reversed(c)) for c in chains]

    symmetry_type_dispatch = {}
    def _symmetry_dispatch(self):
        return dict({
            (3,):self._3_coords,
            (2,1):self._2_1_coords,
            (3,1):self._3_1_coords
        })
    def _2_1_coords(self, atom, neighbors, X, R, backbone=None):
        coords = []
        R = R[0]
        chains = self.get_precedent_chains(R, 1, ignored=[atom] + neighbors, backbone=backbone)
        if len(chains) == 0: chains = [[]]
        for c in chains:
            coords.extend(self.RYX2_coords(c + [R], atom, X))
        return coords
    def _3_coords(self, atom, neighbors, X, backbone=None):
        return self.RYX3_coords([], atom, X)
    def _3_1_coords(self, atom, neighbors, X, R, backbone=None):
        coords = []
        R = R[0]
        chains = self.get_precedent_chains(R, 1, ignored=[atom] + neighbors, backbone=backbone)
        if len(chains) == 0: chains = [[]]
        for c in chains:
            coords.extend(self.RYX3_coords(c + [R], atom, X))
        return coords

    # def _2_2_coords(self, atom, neighbors, X, R, backbone=None):
    #     coords = []
    #     R = R[0]
    #     # chains = self.get_precedent_chains(R, 1, ignored=[atom] + neighbors, backbone=backbone)
    #     if len(chains) == 0: chains = [[]]
    #     for c in chains:
    #         coords.extend(self.R2YX2_coords(c + [R], atom, X))
    #     return coords


    @classmethod
    def get_symmetry_groups(cls, neighbors, matches):
        groups = {}
        n = len(neighbors)
        k = 0
        for i in range(n):
            if i not in groups:
                groups[i] = {i}
            for j in range(i+1, n):
                eq = matches[k]
                k += 1
                if eq:
                    groups[j] = groups[i]
                    groups[i].add(j)
        groups = {id(g):g for g in groups.values()}
        groups = list(reversed(sorted(groups.values(), key=lambda g:len(g))))
        return [
            [neighbors[i] for i in g]
            for g in groups
        ]

    def symmetry_coords(self, atom, neighborhood=3, backbone=None):
        # neighbors = list(self.graph.map[atom])
        coords = []
        # dispatch = self._symmetry_dispatch()
        # neighbor_counts = {sum(k) for k in dispatch.keys()}
        # if len(neighbors) in neighbor_counts:
        #     symms = self.get_neighborhood_symmetries(neighbors, ignored=[atom], neighborhood=neighborhood)
        #     groups = self.get_symmetry_groups(neighbors, symms)
        #     key = tuple(len(g) for g in groups)
        #     dfunc = dispatch.get(key)
        #     if dfunc is not None:
        #         coords = dfunc(atom, neighbors, *groups, backbone=backbone)
        chains = self.get_precedent_chains(atom, 3, backbone=backbone)
        if len(chains) == 0: chains = [[]]
        for R in chains:
            coords.extend(self.chain_coords(R, atom))

        # if coords is None:
        #     R = self.get_precedent_chain(atom, 3, backbone=backbone)
        #     coords = self.chain_coords(R, atom)

        return coords


"""
Sets up a general Interpolator class that looks like Mathematica's InterpolatingFunction class
"""
import typing

import numpy as np, abc, enum, math
import scipy.interpolate as interpolate
import scipy.spatial as spat
from .Mesh import Mesh, MeshType

from .. import Numputils as nput
from .NeighborBasedInterpolators import *

__all__ = [
    "Interpolator",
    "Extrapolator",
    "RBFDInterpolator",
    "InverseDistanceWeightedInterpolator",
    "ProductGridInterpolator",
    "UnstructuredGridInterpolator",
    "CoordinateInterpolator"
]


class InterpolatorException(Exception):
    pass


######################################################################################################
##
##                                   Interpolator Class
##
######################################################################################################

class BasicInterpolator(metaclass=abc.ABCMeta):
    """
    Defines the abstract interface we'll need interpolator instances to satisfy so that we can use
    `Interpolator` as a calling layer
    """

    @abc.abstractmethod
    def __init__(self, grid, values, **opts):
        raise NotImplementedError("abstract interface")

    @abc.abstractmethod
    def __call__(self, points, **kwargs):
        raise NotImplementedError("abstract interface")

    @abc.abstractmethod
    def derivative(self, order):
        """
        Constructs the derivatives of the interpolator at the given order
        :param order:
        :type order:
        :return:
        :rtype: BasicInterpolator
        """
        raise NotImplementedError("abstract interface")

class ProductGridInterpolator(BasicInterpolator):
    """
    A set of interpolators that support interpolation
    on a regular (tensor product) grid
    """

    def __init__(self, grids, vals,
                 caller=None, order=None,
                 extrapolate=True,
                 periodic=False,
                 boundary_conditions=None
                 ):
        #TODO: add richer support for different types of boundary conditions to allow us to
        #      handler extrapolations easier, e.g. we can fix the linear term in the first derivative
        #      at the boundary or we can specify that a function is periodic along a given coordinate
        """
        :param grids:
        :type grids:
        :param points:
        :type points:
        :param caller:
        :type caller:
        :param order:
        :type order: int | Iterable[int]
        """

        if order is None:
            order = 3

        if nput.is_numeric(grids[0]):
            grids = [grids]
        self.grids = grids
        self.vals = vals
        self.periodic = periodic
        if caller is None:
            ndim = len(grids)
            if ndim == 1:
                opts = {}
                if order is not None:
                    opts["k"] = order
                caller = self.get_base_spline(
                    grids[0], vals, order,
                    periodic=periodic,
                    boundary_conditions=boundary_conditions,
                    extrapolate=extrapolate
                )
            else:
                caller = self.construct_ndspline(grids, vals, order, extrapolate=extrapolate,
                                                 periodic=periodic,
                                                 boundary_conditions=boundary_conditions
                                                 )
        self.caller = caller

    @classmethod
    def get_base_spline(cls, grid, vals, order, periodic=False, boundary_conditions=None, extrapolate=False):
        base_spline = interpolate.make_interp_spline(
            grid, vals, k=order,
            bc_type='periodic' if periodic else boundary_conditions
        )

        # Adapated from interpolate.PPoly.from_spline

        if vals.ndim > 1:
            tck = base_spline.tck
            t, c, k = tck
            polys = [interpolate.PPoly.from_spline((t, cj, k)) for cj in c.T]
            cc = np.dstack([p.c for p in polys])
            return interpolate.PPoly.construct_fast(cc, polys[0].x, extrapolate)
        else:
            return interpolate.PPoly.from_spline(base_spline.tck, extrapolate)

    @classmethod
    def construct_ndspline(cls, grids, vals, order, extrapolate=True,
                           periodic=False,
                           boundary_conditions=None
                           ):
        """
        Builds a tensor product ndspline by constructing a product of 1D splines

        :param grids: grids for each dimension independently
        :type grids: Iterable[np.ndarray]
        :param vals:
        :type vals: np.ndarray
        :param order:
        :type order: int | Iterable[int]
        :return:
        :rtype: interpolate.NdPPoly
        """

        # inspired by `csaps` python package
        # idea is to build a spline approximation in
        # every direction (where the return values are multidimensional)
        ndim = len(grids)
        if isinstance(order, (int, np.integer)):
            order = [order] * ndim

        if periodic is True or periodic is False:
            periodic = [periodic] * ndim
        if (
                boundary_conditions is None
                or isinstance(boundary_conditions, str)
                or len(boundary_conditions[0]) == 2 and nput.is_numeric(boundary_conditions[0][0])
        ):
            boundary_conditions = [boundary_conditions] * ndim

        coeffs = vals
        x = [None]*ndim
        for i, (g, o, p, bcs) in enumerate(zip(grids, order, periodic, boundary_conditions)):
            og_shape = coeffs.shape
            coeffs = coeffs.reshape((len(g), -1)).T
            sub_coeffs = [np.empty(0)]*len(coeffs)
            for e,v in enumerate(coeffs):
                ppoly = cls.get_base_spline(g, v, o, periodic=p, boundary_conditions=bcs)
                x[i] = ppoly.x
                sub_coeffs[e] = ppoly.c
            coeffs = np.array(sub_coeffs)
            coeffs = coeffs.reshape(
                og_shape[1:]+
                    sub_coeffs[0].shape
            )
            tot_dim = ndim+i+1
            coeffs = np.moveaxis(coeffs, tot_dim-2, tot_dim-2-i)

        return interpolate.NdPPoly(coeffs, x, extrapolate=extrapolate)

    def handle_periodicity(self, coords):
        if not self.periodic: return coords

        coords = np.asanyarray(coords)
        ndim = len(self.grids)
        one_d = coords.ndim == 0
        if one_d:
            coords = coords[np.newaxis]
        smol = coords.ndim == 1
        if smol:
            if ndim == 1:
                coords = coords[:, np.newaxis]
            else:
                coords = coords[np.newaxis]

        periodic = self.periodic
        if periodic is True: periodic = [periodic] * ndim
        new_coords = coords.copy()
        for i,(g,p) in enumerate(zip(self.grids, periodic)):
            if p:
                ptp = g[-1] - g[0]
                new_coords[..., i] = np.mod((coords[..., i] - g[0]), ptp) + g[0]

        if smol:
            if ndim == 1:
                new_coords = new_coords[:, 0]
            else:
                new_coords = new_coords[0]

        if one_d:
            new_coords = new_coords[0]

        return new_coords


    def __call__(self, coords, *etc, **kwargs):
        """
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype: np.ndarray
        """
        coords = self.handle_periodicity(coords)
        return self.caller(coords, *etc, **kwargs)

    def derivative(self, order):
        """
        :param order:
        :type order:
        :return:
        :rtype: ProductGridInterpolator
        """
        # ndim = len(self.grids)
        # if ndim == 1:
        return type(self)(
            self.grids,
            self.vals,
            caller=self.caller.derivative(order),
            periodic=self.periodic
        )
        # elif ndim == 2:
        #     def caller(coords, _og=self.caller, **kwargs):
        #         return _og(coords, dx=order[0], dy=order[1], **kwargs)
        #     return type(self)(
        #         self.grids,
        #         self.vals,
        #         caller=caller
        #     )
        # else:
        #     derivs = self.caller.derivative(order)
        #     raise NotImplementedError("woof")

class UnstructuredGridInterpolator(BasicInterpolator):
    """
    Defines an interpolator appropriate for totally unstructured grids by
    delegating to the scipy `RBF` interpolators
    """

    default_neighbors=25
    def __init__(self, grid, values, order=None, neighbors=None, extrapolate=True, **opts):
        """
        :param grid:
        :type grid: np.ndarray
        :param values:
        :type values:  np.ndarray
        :param order:
        :type order: int
        :param neighbors:
        :type neighbors: int
        :param extrapolate:
        :type extrapolate: bool
        :param opts:
        :type opts:
        """
        self.extrapolate=extrapolate
        self._hull = None
        self._grid = grid

        if neighbors is None:
            neighbors = np.min([self.default_neighbors, len(grid)])

        if order is not None:
            if isinstance(order, int):
                if order == 1:
                    order = "linear"
                elif order == 3:
                    order = "cubic"
                elif order == 5:
                    order = "quintic"
                else:
                    raise InterpolatorException("{} doesn't support interpolation order '{}'".format(
                        interpolate.RBFInterpolator,
                        order
                    ))
            self.caller = interpolate.RBFInterpolator(grid, values, kernel=order, neighbors=neighbors, **opts)
        else:
            self.caller = interpolate.RBFInterpolator(grid, values, neighbors=neighbors, **opts)

    def _member_q(self, points):
        """
        Checks if the points are in the interpolated convex hull
        in the case that we aren't extrpolating so we can return
        NaN for those points
        :param points:
        :type points:
        :return:
        :rtype:
        """
        if self._hull is None:
            self._hull = spat.ConvexHull(self._grid)
            self._hull = spat.Delaunay(self._hull)
        return self._hull.find_simplex(points) >= 0
    def __call__(self, points):
        if self.extrapolate:
            return self.caller(points)
        else:
            hull_points = self._member_q(points)
            res = np.full(len(points), np.nan)
            res[hull_points] = self.caller(points[hull_points])
            return res

    def derivative(self, order):
        """
        Constructs the derivatives of the interpolator at the given order
        :param order:
        :type order:
        :return:
        :rtype: UnstructuredGridInterpolator
        """
        raise NotImplementedError("derivatives not implemented for unstructured grids")

# class _PolyFitManager:
#     """
#     Provides ways to evaluate polynomials
#     fast and get matrices of terms to fit
#     """
#     def __init__(self, order, ndim):
#         self.order = order
#         self.ndim = ndim
#         self._expr = None
#     @property
#     def perms(self):
#         from ..Combinatorics import SymmetricGroupGenerator
#
#         if self._perms is None:
#             # move this off to cached classmethod
#             # since there are restricted numbers of (order, ndim) pairs
#             sn = SymmetricGroupGenerator(self.ndim)
#             self._perms = sn.get_terms(list(range(self.order+1)))
#         return self._perms
#     def eval_poly_mat(self, coords):
#         """
#
#         :param coords: shape=(npts, ndim)
#         :type coords: np.ndarray
#         :return:
#         :rtype:
#         """
#         return np.power(coords[:, np.newaxis, :], self.perms[np.newaxis])
#     def eval_poly_mat_deriv(self, which):
#         which, counts = np.unique(which, return_counts=True)
#         perm_inds = self.perms[:, which]
#         shift_perms = perm_inds - counts[np.newaxis, :] # new monom orders


class ExtrapolatorType(enum.Enum):
    Default='Automatic'
    Error='Raise'

class Interpolator:
    """
    A general purpose that takes your data and just interpolates it without whining or making you do a pile of extra work
    """
    DefaultExtrapolator = ExtrapolatorType.Default
    def __init__(self,
                 grid,
                 vals,
                 interpolation_function=None,
                 interpolation_order=None,
                 extrapolator=None,
                 extrapolation_order=None,
                 **interpolation_opts
                 ):
        """
        :param grid: an unstructured grid of points **or** a structured grid of points **or** a 1D array
        :type grid: np.ndarray
        :param vals: the values at the grid points
        :type vals: np.ndarray
        :param interpolation_function: the basic function to be used to handle the raw interpolation
        :type interpolation_function: None | BasicInterpolator
        :param interpolation_order: the order of extrapolation to use (when applicable)
        :type interpolation_order: int | str | None
        :param extrapolator: the extrapolator to use for data points not on the grid
        :type extrapolator: Extrapolator | None | str | function
        :param extrapolation_order: the order of extrapolation to use by default
        :type extrapolation_order: int | str | None
        :param interpolation_opts: the options to be fed into the interpolating_function
        :type interpolation_opts:
        """
        self.grid = grid = Mesh(grid) if not isinstance(grid, Mesh) else grid
        self.vals = vals
        if interpolation_function is None:
            interpolation_function = self.get_interpolator(grid, vals,
                                                           interpolation_order=interpolation_order,
                                                           allow_extrapolation=extrapolator is None,
                                                           **interpolation_opts
                                                           )
        self.interpolator = interpolation_function

        if extrapolator is not None:
            if isinstance(extrapolator, ExtrapolatorType):
                if extrapolator == ExtrapolatorType.Default:
                    extrapolator = self.get_extrapolator(grid, vals, extrapolation_order=extrapolation_order)
                else:
                    raise ValueError("don't know what do with extrapolator type {}".format(extrapolator))
            elif not isinstance(extrapolator, Extrapolator):
                extrapolator = Extrapolator(extrapolator)
        self.extrapolator = extrapolator

    @classmethod
    def get_interpolator(cls, grid, vals, interpolation_order=None, allow_extrapolation=True, **opts):
        """Returns a function that can be called on grid points to interpolate them

        :param grid:
        :type grid: Mesh
        :param vals:
        :type vals: np.ndarray
        :param interpolation_order:
        :type interpolation_order: int | str | None
        :param opts:
        :type opts:
        :return: interpolator
        :rtype: function
        """
        if grid.ndim == 1:
            interpolator = ProductGridInterpolator(
                grid,
                vals,
                order=interpolation_order,
                extrapolate=allow_extrapolation
            )
        elif (
                grid.mesh_type == MeshType.Structured
                or grid.mesh_type == MeshType.Regular
        ):
            interpolator = ProductGridInterpolator(
                grid.subgrids,
                vals,
                order=interpolation_order,
                extrapolate=allow_extrapolation
            )
        elif grid.mesh_type == MeshType.Unstructured:
            # for now we'll only use the RadialBasisFunction interpolator, but this may be extended in the future
            interpolator = UnstructuredGridInterpolator(
                grid,
                vals,
                order=interpolation_order,
                extrapolate=allow_extrapolation
            )
        elif grid.mesh_type == MeshType.SemiStructured:
            raise NotImplementedError("don't know what I want to do with semistructured meshes anymore")
        else:
            raise InterpolatorException("{}.{}: can't handle mesh_type '{}'".format(
                cls.__name__,
               'get_interpolator',
                grid.mesh_type
            ))

        return interpolator

    @classmethod
    def get_extrapolator(cls, grid, vals, extrapolation_order=1, **opts):
        """
        Returns an Extrapolator that can be called on grid points to extrapolate them

        :param grid:
        :type grid: Mesh
        :param extrapolation_order:
        :type extrapolation_order: int
        :return: extrapolator
        :rtype: Extrapolator
        """

        # Extrapolator(
        #     cls(
        #         grid,
        #         vals,
        #         interpolation_order=extrapolation_order,
        #         extrapolator=Extrapolator(lambda g: np.full(g.shape, np.nan))
        #     )
        # )

        if grid.ndim == 1:
            extrapolator = ProductGridInterpolator(
                grid,
                vals,
                order=extrapolation_order,
                extrapolate=True
            )
        elif (
                grid.mesh_type == MeshType.Structured
                or grid.mesh_type == MeshType.Regular
        ):
            extrapolator = ProductGridInterpolator(
                grid.subgrids,
                vals,
                order=extrapolation_order,
                extrapolate=True
            )
        elif grid.mesh_type == MeshType.Unstructured:
            # for now we'll only use the RadialBasisFunction interpolator, but this may be extended in the future
            extrapolator = UnstructuredGridInterpolator(
                grid,
                vals,
                neighbors=extrapolation_order+1,
                order=extrapolation_order,
                extrapolate=True
            )
        elif grid.mesh_type == MeshType.SemiStructured:
            raise NotImplementedError("don't know what I want to do with semistructured meshes anymore")
        else:
            raise InterpolatorException("{}.{}: can't handle mesh_type '{}'".format(
                cls.__name__,
               'get_interpolator',
                grid.mesh_type
            ))

        return Extrapolator(
            extrapolator,
            **opts
        )

    def apply(self, grid_points, **opts):
        """Interpolates then extrapolates the function at the grid_points

        :param grid_points:
        :type grid_points:
        :return:
        :rtype:
        """
        # determining what points are "inside" an interpolator region is quite tough
        # instead it is probably better to allow the basic thing to interpolate and do its thing
        # and then allow the extrapolator to post-process that result
        vals = self.interpolator(grid_points, **opts)
        if self.extrapolator is not None:
            vals = self.extrapolator(grid_points, vals)
        return vals

    def derivative(self, order):
        """
        Returns a new function representing the requested derivative
        of the current interpolator

        :param order:
        :type order:
        :return:
        :rtype:
        """
        return type(self)(
            self.grid,
            self.vals,
            interpolation_function=self.interpolator.derivative(order),
            extrapolator=self.extrapolator.derivative(order) if self.extrapolator is not None else self.extrapolator
        )

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

######################################################################################################
##
##                                   Extrapolator Class
##
######################################################################################################
class Extrapolator:
    """
    A general purpose that takes your data and just extrapolates it.
    This currently only exists in template format.
    """
    def __init__(self,
                 extrapolation_function,
                 warning=False,
                 **opts
                 ):
        """
        :param extrapolation_function: the function to handle extrapolation off the interpolation grid
        :type extrapolation_function: None | function | Callable | Interpolator
        :param warning: whether to emit a message warning about extrapolation occurring
        :type warning: bool
        :param opts: the options to feed into the extrapolator call
        :type opts:
        """
        self.extrapolator = extrapolation_function
        self.extrap_warning = warning
        self.opts = opts

    def derivative(self, n):
        return type(self)(
            self.extrapolator.derivative(n),
            warning=self.extrapolator
        )

    def find_extrapolated_points(self, gps, vals, extrap_value=np.nan):
        """
        Currently super rough heuristics to determine at which points we need to extrapolate
        :param gps:
        :type gps:
        :param vals:
        :type vals:
        :return:
        :rtype:
        """
        if extrap_value is np.nan:
            where = np.isnan(vals)
        elif extrap_value is np.inf:
            where = np.isinf(vals)
        elif not isinstance(extrap_value, (int, float, np.floating, np.integer)):
            where = np.logical_and(vals <= extrap_value[0], vals >= extrap_value[1])
        else:
            where = np.where(vals == extrap_value)

        return gps[where], where

    def apply(self, gps, vals, extrap_value=np.nan):
        ext_gps, inds = self.find_extrapolated_points(gps, vals, extrap_value=extrap_value)
        if len(ext_gps) > 0:
            new_vals = self.extrapolator(ext_gps)
            # TODO: emit a warning about extrapolating if we're doing so?
            vals[inds] = new_vals
        return vals

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

class IncrementalCartesianCoordinateInterpolation:
    def __init__(self,
                 abcissae,
                 coords,
                 *,
                 coordinate_system,
                 max_displacement_step=1.0,
                 max_refinements=1,
                 reembed=False,
                 embedding_options=None
                 ):
        self.abcissae = list(abcissae)
        self.coords = self.prep_cartesians(coords)
        self.converter = self.prep_coordinate_system_converter(coordinate_system)
        self.internals = [
            self.converter(c) for c in self.coords
        ]
        self.max_disp_step = max_displacement_step
        self.max_refinements = max_refinements
        self.reembed=reembed
        self.embedding_options={} if embedding_options is None else embedding_options

    @classmethod
    def wrap_convert(cls, system):
        def convert(coords):
            return coords.convert(system)
        return convert

    @classmethod
    def prep_coordinate_system_converter(cls, coordinate_system):
        from McUtils.Coordinerds import CartesianCoordinates3D

        if coordinate_system is None:
            return cls.wrap_convert(CartesianCoordinates3D)
        elif hasattr('coordinate_system', 'convert'):
            return cls.wrap_convert(coordinate_system)
        else:
            return coordinate_system

    @classmethod
    def refined_step_conv(cls,
                          pct,
                          converter,
                          init_abc, final_abc,
                          init_coords, final_coords,
                          init_internals, final_internals,
                          max_refinements=None, max_disp=.5,
                          reembed=False, embedding_options=None
                          ):
        from McUtils.Coordinerds import CoordinateSet

        if embedding_options is None:
            embedding_options = {}

        # d = CoordinateSet(self.internals[start] + disp, self.internals[start].system)
        # d.convert(self.coords[start].system)
        if pct > .5:
            return cls.refined_step_conv(
                1 - pct,
                converter,
                final_abc, init_abc,
                final_coords, init_coords,
                final_internals, init_internals,
                max_refinements=max_refinements, max_disp=max_disp,
                reembed=reembed
            )

        disp = final_coords.convert(init_internals.system) - init_internals
        step = final_abc - init_abc

        new_abcissae = []
        new_internals = []
        new_coords = []

        d = pct * disp
        md = np.max(np.abs(d))
        n_ref = 0
        while md > max_disp and (max_refinements is None or n_ref < max_refinements):
            scaling = max_disp / md
            d = d * scaling
            new_abc = init_abc + (step * pct * scaling)
            new_step = final_abc - new_abc
            pct = pct * (1 - scaling) / (1 - pct * scaling) # make sure we end up at the same spot
            init_abc = new_abc
            step = new_step
            target_internals = init_internals + d
            new_disp = CoordinateSet(target_internals, init_internals.system)
            # new_disp.converter_options = init_internals.system.converter_options
            new_carts = new_disp.convert(init_coords.system)
            if reembed:
                emb = nput.eckart_embedding(
                    init_coords,
                    new_carts,
                    **embedding_options
                )
                new_carts = (new_carts*0) + emb.coordinates
            init_coords = new_carts
            init_internals = converter(init_coords)
            new_abcissae.append(init_abc)
            new_coords.append(init_coords)
            new_internals.append(init_internals)

            disp = final_coords.convert(init_internals.system) - init_internals
            d = pct * disp
            md = np.max(np.abs(d))
            n_ref += 1

        target_internals = CoordinateSet(init_internals + d, init_internals.system)
        new_carts = target_internals.convert(init_coords.system)
        if reembed:
            emb = nput.eckart_embedding(
                init_coords,
                new_carts,
                **embedding_options
            )
            new_carts = (new_carts*0) + emb.coordinates
        return new_carts, (new_abcissae, new_coords, new_internals)

    def prep_cartesians(self, coords):
        from ..Coordinerds import CoordinateSet, CartesianCoordinates3D
        coords = np.asanyarray(coords)
        if not hasattr(coords, 'system'):
            coords = CoordinateSet(coords, CartesianCoordinates3D)
        return list(coords)

    def incremental_interp(self, start, point):
        b = self.abcissae[start]
        a = self.abcissae[start+1]
        width = a - b
        pct = (point - b) / width

        coords, (new_abcissae, new_coords, new_internals) = self.refined_step_conv(
            pct,
            self.converter,
            b, a,
            self.coords[start], self.coords[start+1],
            self.internals[start], self.internals[start+1],
            max_refinements=self.max_refinements,
            max_disp=self.max_disp_step,
            reembed=self.reembed
        )

        self.abcissae = self.abcissae[:start+1] + new_abcissae + self.abcissae[start+1:]
        self.coords = self.coords[:start+1] + new_coords + self.coords[start+1:]
        self.internals = self.internals[:start+1] + new_internals + self.internals[start+1:]

        return coords

    def interpolate(self, point):
        point = np.asanyarray(point)
        base_shape = point.shape
        point = point.reshape(-1)
        ord = np.argsort(point)
        point = point[ord]

        vals = np.empty(point.shape + self.coords[0].shape, dtype=self.coords[0].dtype)
        for k,p in enumerate(point):
            i = np.searchsorted(self.abcissae, p)
            if i > 0: i = i - 1
            crds = self.incremental_interp(i, p)
            vals[k] = crds
        inv = np.argsort(ord)
        vals = vals[inv,]
        return vals.reshape(base_shape + vals.shape[-2:])

    def __call__(self, point):
        return self.interpolate(point)

class CoordinateInterpolator:

    default_interpolator_type = IncrementalCartesianCoordinateInterpolation
    # default_smoothed_interpolator_type = IncrementalCartesianCoordinateInterpolation
    def __init__(self,
                 coordinates,
                 arc_lengths=None,
                 distance_function=None,
                 base_interpolator=None,
                 coordinate_system=None,
                 **interpolator_options
                 ):
        coordinates = np.asanyarray(coordinates)

        self.distance_function, abcissae = self.get_arc_lengths(
            coordinates,
            arc_lengths,
            distance_function=distance_function
        )

        if base_interpolator is None:
            interpolator_options['coordinate_system'] = interpolator_options.get(
                'coordinate_system',
                coordinate_system
            )
            base_interpolator = self.default_interpolator_type

        self.interpolator = base_interpolator(
            abcissae,
            coordinates,
            **interpolator_options
        )

    @classmethod
    def euclidean_coordinate_distance(cls, p1, p2):
        return np.linalg.norm(p2 - p1)


    @classmethod
    def lookup_distance_function(cls, distance_function):
        return {
            'uniform':cls.uniform_distance_function
        }[distance_function.lower()]

    @classmethod
    def uniform_distance_function(cls, coords):
        return np.linspace(0, 1, len(coords))

    @classmethod
    def get_arc_lengths(cls,
                        coordinates:np.ndarray,
                        arc_lengths=None,
                        distance_function:'typing.Callable[[np.ndarray, np.ndarray], float]'=None
                        ):

        if arc_lengths is None:
            if isinstance(distance_function, str):
                arc_lengths = cls.lookup_distance_function(distance_function)(coordinates)
            else:
                if distance_function is None:
                    distance_function = cls.euclidean_coordinate_distance
                arc_lengths = np.cumsum([0.] + [
                    distance_function(p1, p2)
                    for p1, p2 in zip(coordinates[:-1], coordinates[1:])
                ])
        arc_lengths = np.asanyarray(arc_lengths)

        return distance_function, (arc_lengths - np.min(arc_lengths)) / np.ptp(arc_lengths)

    def __call__(self, points, **etc):
        return self.interpolator(points, **etc)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) b3m2a1.
# Distributed under the terms of the Modified BSD License.

import pytest

from ..example import ExampleWidget


def test_example_creation_blank():
    w = ExampleWidget()
    assert w.value == 'Hello World'



"""
Provides minimal utilities for making slurm stuff nicer
"""

import inspect

__all__ = ["SBatchJob"]

class SBatchJob:
    """
    Provides a simple interface to formatting SLURM
    files so that they can be submitted to `sbatch`.
    The hope is that this can be subclassed codify
    options for different HPC paritions and whatnot.
    """
    slurm_keys = [
        "account", "acctg-freq", "array", "batch", "bb", "bbf", "begin", "chdir",
        "cluster-constraint", "clusters", "comment", "constraint", "contiguous",
        "core-spec", "cores-per-socket", "cpu-freq", "cpus-per-gpu", "cpus-per-task",
        "deadline", "delay-boot", "dependency", "distribution", "error", "exclude",
        "exclusive", "export", "export-file", "extra-node-info",
        "get-user-env", "gid", "gpu-bind", "gpu-freq", "gpus", "gpus-per-node",
        "gpus-per-socket", "gpus-per-task", "gres", "gres-flags", "help", "hint",
        "ignore-pbs", "input", "job-name", "kill-on-invalid-dep", "licenses",
        "mail-type", "mail-user", "mcs-label", "mem", "mem-bind", "mem-per-cpu",
        "mem-per-gpu", "mincpus", "network", "nice", "nodefile", "nodelist", "nodes",
        "no-kill", "no-requeue", "ntasks", "ntasks-per-core", "ntasks-per-gpu",
        "ntasks-per-node", "ntasks-per-socket", "open-mode", "output", "overcommit",
        "oversubscribe", "parsable", "partition", "power", "priority", "profile",
        "propagate", "qos", "quiet", "reboot", "requeue", "reservation", "signal",
        "sockets-per-node", "spread-job", "switches", "test-only", "thread-spec",
        "threads-per-core", "time", "time-min", "tmp", "uid", "usage", "use-min-nodes",
        "verbose", "version", "wait", "wait-all-nodes", "wckey", "wrap"
    ]
    default_opts = {
        'chdir': "."
    }
    def __init__(self,
                 description=None,
                 job_name=None, account=None, partition=None,
                 mem=None,  nodes=None, ntasks_per_node=None,
                 chdir=None, output=None,
                 steps=(),
                 **opts
                 ):
        self.description=description
        self.steps=steps

        base_opts = dict(
            job_name=job_name, account=account, partition=partition,
            mem=mem, nodes=nodes, ntasks_per_node=ntasks_per_node,
            output=output, chdir=chdir
        )
        base_opts = self.clean_opts(base_opts)
        base_opts = dict(self.default_opts, **base_opts)

        opts = dict(opts, **base_opts)
        self.opts = self.clean_opts(opts)

    def clean_opts(self, opts):
        """
        Makes sure opt names are clean.
        Does no validation of the values sent in.

        :param opts:
        :type opts:
        :return:
        :rtype:
        """
        clean_opts = {}
        for opt_name, opt_val in opts.items():
            opt_name = opt_name.replace("_", "-")
            if opt_val is not None:
                if opt_name not in self.slurm_keys:
                    raise ValueError("SBATCH option {} invalid; accepted ones are {}".format(opt_name, self.slurm_keys))
                clean_opts[opt_name] = opt_val
        return clean_opts

    sbatch_opt_template="#SBATCH --{name}={value}"
    def format_opt_block(self):
        """
        Formats block of options
        :return:
        :rtype:
        """
        return "\n".join(
            self.sbatch_opt_template.format(name=k, value=v) for k,v in self.opts.items() if v is not None
        )

    sbatch_template="#!/bin/bash\n{opts}\n\n{enter}\n{call}\n{exit}"
    sbatch_enter_command="\n".join([
        'echo "Starting Job $SLURM_JOB_NAME"',
        'START=$(date +%s.%N)',
        'echo "  START: $(date)"',
        'echo "    PWD: $PWD"',
        'echo "  NODES: $SLURM_JOB_NUM_NODES"',
        'echo "  PART.: $SLURM_JOB_PARTITION"',
        'echo "{sep}"'.format(sep="="*50)
        ])
    sbatch_exit_command = "\n".join([
        'echo "{sep}"'.format(sep="=" * 50),
        'END=$(date +%s.%N)',
        'DIFF=$(echo "$END - $START" | bc)',
        'echo "   END: $(date)"',
        'echo "  TIME: $DIFF"'
    ])
    def format(self):
        """
        Formats an SBATCH file from the held options
        :param call_steps:
        :type call_steps:
        :return:
        :rtype:
        """
        opts = self.format_opt_block()
        enter = self.sbatch_enter_command
        if self.description is not None:
            enter += (
                    '\necho "'
                    + inspect.cleandoc(self.description).replace("\n", '"\necho "')
                    + '"\necho\n\n'
            )
        exit = self.sbatch_exit_command

        call = "\n".join(self.steps)

        return self.sbatch_template.format(
            opts=opts,
            call=call,
            enter=enter,
            exit=exit
        )


import collections
import itertools

import numpy as np
# import scipy.spatial
from ... import Devutils as dev
from ... import Numputils as nput
from ...Data import AtomData, UnitsData
from ...ExternalPrograms import Open3DInterface as o3d

__all__ = [
    "SphereUnionSurface",
    "SphereUnionSurfaceMesh"
]

# class TriangleMesh:
#
#     ...

class SphereUnionSurface:

    default_samples = 50
    default_scaling = 1
    default_expansion = 0
    default_tolerance = 1e-3
    def __init__(self, centers, radii, scaling=None, expansion=None, samples=None, tolerance=None):
        self.centers = np.asanyarray(centers)
        self.radii = np.asanyarray(radii)
        if samples is None:
            samples = self.default_samples
        self.samples = samples
        if scaling is None:
            scaling = self.default_scaling
        self.scaling = scaling
        if expansion is None:
            expansion = self.default_expansion
        if tolerance is None:
            tolerance = self.default_tolerance
        self.tolerance = tolerance
        self.expansion = expansion
        self._sample_points = None

    @classmethod
    def from_xyz(cls,
                 atoms, positions,
                 scaling=None, expansion=None, samples=None,
                 tolerance=None,
                 radius_property='IconRadius',
                 distance_units='BohrRadius'):
        radii = np.array([
            AtomData[a, radius_property] * UnitsData.convert("Angstroms", distance_units)
            for a in atoms
        ])

        return cls(positions, radii,
                   samples=samples, scaling=scaling, expansion=expansion,
                   tolerance=tolerance
                   )

    @property
    def sampling_points(self):
        if self._sample_points is None:
            self._sample_points = self.generate_points()
        return self._sample_points
    @sampling_points.setter
    def sampling_points(self, pts):
        if pts is not None: pts = np.asanyarray(pts)
        self._sample_points = pts

    @classmethod
    def nearest_centers(cls, pts, centers, return_normals=False):
        center_vecs = pts[:, np.newaxis, :] - centers[np.newaxis, :, :]
        center_dm = np.linalg.norm(center_vecs, axis=-1)
        np.fill_diagonal(center_dm, 1e6)
        nearest_centers = np.argmin(center_dm, axis=1)
        if return_normals:
            norms = center_dm[np.arange(len(pts)), nearest_centers]
            return nearest_centers, (norms, center_vecs[np.arange(len(pts)), nearest_centers, :] / norms[:, np.newaxis])
        return nearest_centers

    @classmethod
    def sphere_project(cls, pts, centers, radii):
        center_vecs = pts[:, np.newaxis, :] - centers[np.newaxis, :, :]
        center_dm = np.linalg.norm(center_vecs, axis=-1)
        np.fill_diagonal(center_dm, 1e6)
        nearest_centers = np.argmin(center_dm, axis=1)
        sel = (np.arange(len(pts)), nearest_centers)
        scaling = radii[nearest_centers] / center_dm[sel]
        pts = centers[nearest_centers, :] + center_vecs[sel] * scaling[:, np.newaxis]

        return pts

    @classmethod
    def sphere_boundary_pruning(
            cls,
            pts,
            centers,
            # radii,
            min_component=None,
            # max_iterations=15
    ):
        center_vecs = pts[:, np.newaxis, :] - centers[np.newaxis, :, :]
        center_dm = np.linalg.norm(center_vecs, axis=-1)
        np.fill_diagonal(center_dm, 1e6)
        nearest_centers = np.argmin(center_dm, axis=1)

        center_inds, subgroups = nput.group_by(pts, nearest_centers)[0]
        subgroups = list(subgroups)
        if min_component is None:
            primary_comps = []
            for group in subgroups:
                dm = nput.distance_matrix(group)
                np.fill_diagonal(dm, 1e6)
                min_vals = np.min(dm, axis=1)
                vals, bins = np.histogram(min_vals, bins=len(group)//5)
                i = np.argmax(vals)
                primary_comps.append(
                    (bins[i] + bins[i + 1]) / 2
                )
            min_component = np.min(primary_comps) #* .5
        for n,(c, group) in enumerate(zip(center_inds, subgroups)):
            dm = nput.distance_matrix(group)
            np.fill_diagonal(dm, 1e6)
            min_vals = np.min(dm, axis=1)
            print("!", np.sort(min_vals))
            for m,(k, group2) in enumerate(zip(center_inds[n+1:], subgroups[n+1:])):
                dists = np.linalg.norm(group2[:, np.newaxis, :] - group[np.newaxis, :, :], axis=-1)
                dist_cutoff = np.where(np.min(dists, axis=1) < min_component)
                if len(dist_cutoff) > 0 or len(dist_cutoff[0]) > 0:
                    subgroups[n+1+m] = np.delete(group2, dist_cutoff[0], axis=0)

        return np.concatenate(subgroups, axis=0)

    @classmethod
    def point_cloud_repulsion(
            cls,
            pts,
            centers,
            radii,
            min_displacement_cutoff=1e-3,
            stochastic_factor=.0001,
            force_constant=.001,
            power=-3,
            max_iterations=15
    ):
        rows, cols = np.triu_indices(len(pts), k=1)
        n = len(pts)
        for i in range(max_iterations):
            d_vecs = pts[rows, :] - pts[cols, :]
            norms = np.linalg.norm(d_vecs, axis=-1)
            forces = force_constant * np.power(norms, power)
            if np.all(forces < min_displacement_cutoff): break
            force_vecs = d_vecs * forces[:, np.newaxis]
            d_mat = np.zeros((n, n, 3))
            d_mat[rows, cols] = force_vecs
            d_mat[cols, rows] = force_vecs
            disps = np.sum(d_mat, axis=1)

            disps = disps + stochastic_factor * nput.vec_normalize(np.random.normal(size=disps.shape))

            center_vecs = pts[:, np.newaxis, :] - centers[np.newaxis, :, :]
            center_dm = np.linalg.norm(center_vecs, axis=-1)
            np.fill_diagonal(center_dm, 1e6)
            nearest_centers = np.argmin(center_dm, axis=1)
            sel = (np.arange(len(pts)), nearest_centers)
            proj_vecs = (center_vecs[sel] / center_dm[sel][:, np.newaxis])
            eye = nput.identity_tensors((len(pts),), 3)
            proj = eye - proj_vecs[:, :, np.newaxis] * proj_vecs[:, np.newaxis, :]
            pts = pts + (proj @ disps[:, :, np.newaxis]).reshape(disps.shape)

            pts = cls.sphere_project(pts, centers, radii)

        return pts

    @classmethod
    def adjust_point_cloud_density(self,
                                   pts,
                                   centers=None,
                                   radii=None,
                                   min_component=None,
                                   min_component_bins=30,
                                   min_component_scaling=.7,
                                   # max_component=None,
                                   same_point_cutoff=1e-6,
                                   max_iterations=15):
        if len(pts) == 1: return pts

        if centers is not None and radii is None or radii is not None and centers is None:
            raise ValueError("to constrain points, need both centers and radii")
        elif centers is not None:
            centers = np.asanyarray(centers)
            radii = np.asanyarray(radii)

        if max_iterations > 0:
            dm = nput.distance_matrix(pts)
            # rows, cols = np.triu_indices_from(dm, k=1)
            max_dist = np.max(dm) * 100
            np.fill_diagonal(dm, max_dist)
            min_pos = np.argmin(dm, axis=1)
            min_vals = dm[np.arange(len(dm)), min_pos]
            if min_component is None:
                vals, bins = np.histogram(min_vals, bins=min_component_bins)
                i = np.argmax(vals)
                min_component = min_component_scaling * (bins[i] + bins[i+1]) / 2

            for i in range(max_iterations):
                small_mask = min_vals < min_component
                bad_pos = np.where(small_mask)

                # if max_component is not None:
                #     big_mask = min_vals > max_component
                #     big_pos = np.where(big_mask)
                # else:
                #     big_mask = None
                #     big_pos = None

                small_done = len(bad_pos) == 0 or len(bad_pos[0]) == 0
                big_done = True
                # big_done = big_mask is None or (len(big_pos) == 0 or len(big_pos[0]) == 0 or len(pts) == 1)
                if (
                        len(pts) == 1
                        or (small_done and big_done)
                ):
                    break

                bad_vals = min_vals[bad_pos]
                dropped_points = np.where(bad_vals < same_point_cutoff)

                bad_rows = bad_pos[0]
                bad_cols = min_pos[bad_pos]
                _, r_pos = np.unique(bad_rows, return_index=True)
                _, c_pos = np.unique(bad_cols[r_pos], return_index=True)
                # only treat each point once per iteration by first dropping dupes
                # and then ensuring terms that appear in the rows don't appear in the cols
                bad_pos = bad_rows[r_pos[c_pos]]

                if len(dropped_points) > 0 and len(dropped_points[0]) > 0:
                    drop_rows = dropped_points[0]
                    drop_cols = min_pos[dropped_points]
                    drop_mask = drop_cols > drop_rows
                    # drop_rows = drop_rows[drop_mask]
                    drop_cols = drop_cols[drop_mask]
                    bad_pos = np.concatenate([drop_cols, bad_pos])

                bad_rows = bad_pos
                bad_cols = min_pos[bad_pos]

                merge_pos = np.unique(np.concatenate([bad_rows, bad_cols]))
                rem_pos = np.setdiff1d(np.arange(len(pts)), merge_pos)
                # dm = dm[np.ix_(rem_pos, rem_pos)]
                min_vals = min_vals[rem_pos,]
                inv_map = np.argsort(np.concatenate([rem_pos, merge_pos]))
                min_pos = inv_map[min_pos][rem_pos,]
                bad_mask = min_pos >= len(rem_pos)
                min_vals[bad_mask] = max_dist
                min_pos[bad_mask] = -1
                rem_pts = pts[rem_pos,]
                new_pts = (pts[bad_rows] + pts[bad_cols]) / 2 # average point positions

                if centers is not None and radii is not None:
                    d_vecs = new_pts[:, np.newaxis, :] - centers[np.newaxis, :, :]
                    center_dm = np.linalg.norm(d_vecs, axis=-1)
                    np.fill_diagonal(center_dm, max_dist)
                    nearest_centers = np.argmin(center_dm, axis=1)
                    sel = (np.arange(len(new_pts)), nearest_centers)
                    scaling = radii[nearest_centers] / center_dm[sel]
                    new_pts = centers[nearest_centers, :] + d_vecs[sel] * scaling[:, np.newaxis]
                # renormalize to distance to nearest center is unchanged


                if len(new_pts) > 1:
                    new_new_dists = nput.distance_matrix(new_pts)
                    np.fill_diagonal(new_new_dists, max_dist)
                    new_min_pos = np.argmin(new_new_dists, axis=1)
                    new_min_vals = new_new_dists[np.arange(len(new_new_dists)), new_min_pos]
                    dropped_points = np.where(new_min_vals < same_point_cutoff)
                    if len(dropped_points) > 0 and len(dropped_points[0]) > 0:
                        drop_rows = dropped_points[0]
                        drop_cols = new_min_pos[dropped_points]
                        drop_mask = drop_cols > drop_rows
                        # drop_rows = drop_rows[drop_mask]
                        drop_cols = drop_cols[drop_mask]
                        rem_new_pts = np.setdiff1d(np.arange(len(new_pts)), drop_cols)
                        new_pts = new_pts[rem_new_pts,]
                        new_min_vals = new_min_vals[rem_new_pts,]
                        new_min_pos = new_min_pos[rem_new_pts,]
                else:
                    new_min_vals = None
                    new_min_pos = None

                new_rem_dists = np.linalg.norm(rem_pts[:, np.newaxis, :] - new_pts[np.newaxis, :, :], axis=-1)
                new_rem_pos = np.argmin(new_rem_dists, axis=1)
                new_rem_mins = new_rem_dists[np.arange(len(new_rem_dists)), new_rem_pos]
                dropped_points = np.where(new_rem_mins < same_point_cutoff)
                if len(dropped_points) > 0 and len(dropped_points[0]) > 0:
                    drop_rows = dropped_points[0]
                    drop_cols = new_rem_pos[dropped_points]
                    drop_mask = drop_cols > drop_rows
                    # drop_rows = drop_rows[drop_mask]
                    drop_cols = drop_cols[drop_mask]
                    rem_new_pts = np.setdiff1d(np.arange(len(new_pts)), drop_cols)
                    new_pts = new_pts[rem_new_pts,]
                    new_min_vals = new_min_vals[rem_new_pts,]
                    new_rem_dists = np.linalg.norm(rem_pts[:, np.newaxis, :] - new_pts[np.newaxis, :, :], axis=-1)
                    new_rem_pos = np.argmin(new_rem_dists, axis=1)
                    new_rem_mins = new_rem_dists[np.arange(len(new_rem_dists)), new_rem_pos]

                min_mask = min_vals > new_rem_mins
                min_vals[min_mask] = new_rem_mins[min_mask]
                min_pos[min_mask] = new_rem_pos[min_mask]

                new_new_pos = np.argmin(new_rem_dists, axis=0)
                new_new_mins = new_rem_dists[new_new_pos, np.arange(new_rem_dists.shape[1])]
                if new_min_vals is None:
                    new_min_vals = new_new_mins
                    new_min_pos = new_new_pos
                else:
                    min_mask = new_min_vals > new_new_mins
                    new_min_vals[min_mask] = new_new_mins[min_mask]
                    new_min_pos[min_mask] = new_new_pos[min_mask]

                pts = np.concatenate([rem_pts, new_pts], axis=0)
                min_vals = np.concatenate([min_vals, new_min_vals])
                min_pos = np.concatenate([min_pos, new_min_pos])

        return pts

    @classmethod
    def get_exterior_points(cls, points, centers, radii, tolerance=0):
        points = np.asanyarray(points)
        centers = np.asanyarray(centers)
        radii = np.asanyarray(radii)

        dvs = np.linalg.norm(
            points[:, np.newaxis, :] - centers[np.newaxis, :, :],
            axis=-1
        )

        mask = np.all(dvs * (1+tolerance) >= radii[np.newaxis], axis=1)
        return mask

    @classmethod
    def get_surface_points(cls,
                           centers,
                           radii,
                           samples=50,
                           scaling=1,
                           expansion=0,
                           preserve_origins=False,
                           tolerance=0
                           ):
        centers = np.asanyarray(centers)
        radii = np.asanyarray(radii) * scaling + expansion

        base_points = cls.sphere_points(
            centers,
            radii,
            samples
        )
        if not preserve_origins:
            base_points = base_points.reshape(-1, 3)
            mask = cls.get_exterior_points(base_points, centers, radii, tolerance=tolerance)

            return base_points[mask,]
        else:
            subpoints = []
            for bp in base_points:
                mask = cls.get_exterior_points(bp, centers, radii, tolerance=tolerance)
                subpoints.append(bp[mask,])

            return subpoints

    def generate_points(self, scaling=None, expansion=None, samples=None, preserve_origins=False, tolerance=None):
        if samples is None: samples = self.samples
        if scaling is None: scaling = self.scaling
        if expansion is None: expansion = self.expansion
        if tolerance is None: tolerance = self.tolerance

        return self.get_surface_points(
            self.centers,
            self.radii,
            scaling=scaling,
            expansion=expansion,
            samples=samples,
            preserve_origins=preserve_origins,
            tolerance=tolerance
        )


    def generate_mesh(self,
                      points=None,
                      normals=None,
                      scaling=None, expansion=None, samples=None,
                      method='poisson',
                      depth=5,
                      **reconstruction_settings):

        if points is None:
            if scaling is None and expansion is None and samples is None:
                points = self.sampling_points
            else:
                points = self.generate_points(scaling=scaling, expansion=expansion, samples=samples)
        if method == 'poisson':
            if normals is None:
                if isinstance(points, list):
                    normal_list = []
                    for c,p in zip(self.centers, points):
                        normal_list.append(nput.vec_normalize(p - c[np.newaxis]))
                    normals = np.concatenate(normal_list)
                else:
                    _, (_, normals) = self.nearest_centers(points, self.centers, return_normals=True)
            geom = o3d.submodule('geometry')
            pcd = geom.PointCloud()
            pcd.points = o3d.submodule('utility').Vector3dVector(np.asanyarray(points).view(np.ndarray))
            pcd.normals = o3d.submodule('utility').Vector3dVector(np.asanyarray(normals).view(np.ndarray))
            mesh, densities = geom.TriangleMesh.create_from_point_cloud_poisson(
                pcd,
                depth=depth,
                **reconstruction_settings
            )
        else:
            raise NotImplementedError("only Open3D Poisson currently supported")

        return SphereUnionSurfaceMesh.from_o3d(mesh, densities, surf=self)
        # return np.array(mesh.vertices), np.array(mesh.triangles)

    @classmethod
    def sphere_points(cls, centers, radii, samples, generator=None):
        centers = np.asanyarray(centers)
        radii = np.asanyarray(radii)

        base_shape = centers.shape[:-2]
        centers = centers.reshape((-1,) + centers.shape[-2:])
        radii = radii.reshape(-1, radii.shape[-1])

        if generator is None:
            generator = cls.fibonacci_sphere
        if nput.is_int(samples):
            base_points = generator(samples)[np.newaxis, np.newaxis]
        else:
            samples = np.asanyarray(samples)
            if samples.ndim == 1:
                base_points = np.array([
                    generator(n)
                    for n in samples
                ])
                if len(samples) == centers.shape[-2]:
                    base_points = base_points[np.newaxis, :, :, :]
                else:
                    base_points = base_points[:, np.newaxis, :, :]
            else:
                samples = samples.reshape(-1, centers.shape[-2])
                base_points = np.array([
                    [
                        generator(n)
                        for n in subsamp
                    ]
                    for subsamp in samples
                ])

        sphere_points = centers[:, :, np.newaxis, :] + base_points * radii[:, :, np.newaxis, np.newaxis]

        return sphere_points.reshape(base_shape + sphere_points.shape[-3:])

    @classmethod
    def fibonacci_sphere(cls, samples):
        phi = np.pi * (np.sqrt(5.) - 1.)  # golden angle in radians
        samps = np.arange(samples)
        y = 1 - (samps / float(samples - 1)) * 2  # y goes from 1 to -1
        radius = np.sqrt(1 - y * y)  # radius at y
        theta = phi * samps  # golden angle increment
        x = np.cos(theta) * radius
        z = np.sin(theta) * radius

        return np.array([x, y, z]).T

    # def get_triangulation(self, *delaunay_kwargs, **delaunay_opts):
    #     return scipy.spatial.Delaunay(
    #         self.sampling_points,
    #         *delaunay_kwargs,
    #         *delaunay_opts
    #     )

    @classmethod
    def sampling_point_surface_area(cls,
                                    centers,
                                    radii,
                                    points=None,
                                    exterior_test=None,
                                    point_generator=None,
                                    generator_args=None,
                                    center_surface_areas=None,
                                    **test_args
                                    ):
        if points is None:
            if generator_args is None:
                generator_args = {}
            if point_generator is None:
                point_generator = cls.sphere_points
            points = point_generator(
                centers,
                radii,
                **generator_args
            )
        if exterior_test is None:
            exterior_test = cls.get_exterior_points

        if center_surface_areas is None:
            center_surface_areas = 4 * np.pi * radii**2

        a = 0
        for ca, center_points in zip(center_surface_areas, points):
            mask = exterior_test(center_points, centers, radii, **test_args)
            a += ca * np.sum(mask).astype(int) / len(mask)

        return a

    @classmethod
    def _trip_q(self, a, b, c, alpha, beta, gamma, e):
        return a * (
            b**2 + c**2 - a**2
            + beta**2 + gamma**2 - 2*alpha**2
            + e*(b**2 - c**2)
        )
    @classmethod
    def _trip_e(self, a, r2, r3):
        return (r2**2 - r3**2) / a**2
    @classmethod
    def _trip_w(self, a, b, c, alpha, beta, gamma):
        A = alpha**2
        B = beta**2
        C = gamma**2
        a = a**2
        b = b**2
        c = c**2
        d = np.linalg.det([
            [0, 1, 1, 1, 1],
            [1, 0, C, B, A],
            [1, C, 0, a, b],
            [1, B, a, 0, c],
            [1, A, b, c, 0],
        ])
        if d > 0:
            return np.sqrt(1 / 2 * d)
        else:
            return d
    @classmethod
    def _trip_s(self, beta, a, c, q1, q3, e1, e3, w):
        ae = a*(1+e1)
        ce = c*(1-e3)
        t1 = np.arctan(ae*w / (beta*q1) )
        t2 = np.arctan(ce*w / (beta*q3) )
        t3 = np.arctan(2*w / q1)
        t4 = np.arctan(2*w / q3)
        s1 = t1 + t2
        if s1 < 0:
            s1 = np.pi + s1
        elif s1 > np.pi:
            s1 = s1 - np.pi
        if t3 < 0:
            t3 = np.pi + t3
        elif t3 > np.pi:
            t3 = t3 - np.pi
        if t4 < 0:
            t4 = np.pi + t4
        elif t4 > np.pi:
            t4 = t4 - np.pi
        return 2*(beta**2) * s1 - beta*(
            ae*t3
            + ce*t4
        )
    @classmethod
    def _trip_t(cls, a, b, c):
        return (a+b+c)*(-a + b + c)*(a-b+c)*(a+b-c)
    @classmethod
    def _trip_p(cls, a, b, c, r1, r2, r3, t):
        A = (b ** 2 - c ** 2 + r2 ** 2 - r3 ** 2) ** 2
        t_abg = np.sqrt(cls._trip_t(a, r2, r3))
        pp = (A + (t + t_abg) ** 2) / (4 * a ** 2) - r1 ** 2
        pm = (A + (t - t_abg) ** 2) / (4 * a ** 2) - r1 ** 2
        # if pp > 0 and pm < 0:
        #     raise ValueError("?", pp, pm)
        return pp, pm
    @classmethod
    def sphere_triple_intersection_area(cls, a, b, c, r1, r2, r3):
        # https://www.tandfonline.com/doi/pdf/10.1080/00268978800100453
        # https://www-tandfonline-com/doi/epdf/10.1080/00268978700102951
        w = cls._trip_w(a, b, c, r1, r2, r3)
        # if abs(w) < 1e-8:
        #     return 0
        if w < -1e-8:
            t2 = cls._trip_t(a, b, c)
            if t2 < 0: return (), 0 # no intersection
            t = np.sqrt(t2)
            p1p, p1m = cls._trip_p(a, b, c, r1, r2, r3, t)
            p2p, p2m = cls._trip_p(b, a, c, r2, r3, r1, t)
            p3p, p3m = cls._trip_p(c, a, b, r3, r1, r2, t)

            if p1p > 0 and p2p > 0 and p3p > 0:
                p1, p2, p3 = p1m, p2m, p3m
            else:
                p1, p2, p3 = p1p, p2p, p3p

            if p1 > 0 and p2 > 0 and p3 > 0:
                return (), 0
            if p1 <= 0 and p2 > 0 and p3 > 0:
                # _, area = cls.sphere_double_intersection_area(a, r2, r3)
                return (1, 2), None
            if p1 > 0 and p2 <= 0 and p3 > 0:
                # _, area = cls.sphere_double_intersection_area(b, r1, r3)
                return (0, 2), None
            if p1 > 0 and p2 > 0 and p3 <= 0:
                # _, area = cls.sphere_double_intersection_area(c, r1, r2)
                return (0, 1), None
            if p1 <= 0 and p2 <= 0 and p3 > 0:
                return (2,), None#-cls.sphere_area(r3)
            if p1 <= 0 and p2 > 0 and p3 <= 0:
                return (1,), None#-cls.sphere_area(r2)
            if p1 > 0 and p2 <= 0 and p3 <= 0:
                return (0,), None#-cls.sphere_area(r1)
            else:
                raise ValueError((p1 > 0,  p2 > 0,  p3 > 0),
                                 p1, p2, p3)
                raise ValueError("don't know what to do here")

        e1 = cls._trip_e(a, r2, r3) #(r2**2 - r3**2) / a**2
        e2 = cls._trip_e(b, r3, r1) #(r3**2 - r1**2) / b**2
        e3 = cls._trip_e(c, r1, r2) #(r1**2 - r2**2) / c**2

        q1 = cls._trip_q(a, b, c, r1, r2, r3, e1)
        q2 = cls._trip_q(b, c, a, r2, r3, r1, e2)
        q3 = cls._trip_q(c, a, b, r3, r1, r2, e3)

        A1 = cls._trip_s(r1, c, b, q3, q2, e3, e2, w)
        A2 = cls._trip_s(r2, a, c, q1, q3, e1, e3, w)
        A3 = cls._trip_s(r3, b, a, q2, q1, e2, e1, w)


        return None, A1 + A2 + A3

    IntersectionCircle = collections.namedtuple("IntersectionCircle",
                                                ["center", "normal", "radius"]
                                                )
    @classmethod
    def sphere_double_intersection_circle(cls, centers, radii, dist=None):
        d = centers[1] - centers[0]
        a = dist
        u, a = nput.vec_normalize(d, norms=dist, return_norms=True)

        r, R = radii**2
        x = (a**2 - R + r) / (2 * a)
        r = np.sqrt(r - x**2)
        return cls.IntersectionCircle(centers[0] + u*x, u, r)
    @classmethod
    def sphere_triple_intersection_point(cls, centers, radii, dists=None):
        # we assume it has been checked that the three spheres intersect using the determinant
        # above

        #TODO: vectorize for multiple sets of intersections
        #TODO: move into Numputils for axis system finding
        dxy = centers[1:] - centers[(0,)]
        (x, y), norms = nput.vec_normalize(dxy, norms=dists, return_norms=True)
        cxy = np.dot(y, x)
        #TODO: handle colinearity
        y = (y - cxy*x) / np.sqrt(1-cxy**2) # norm known analytically
        z = nput.vec_crosses(x, y)# normalized by construction
        axes = np.array([x, y, z])

        #TODO: rederive these w.r.t cos terms and norms...
        cxy = np.dot(x, dxy[1])
        cyy = np.dot(y, dxy[1])

        R1, R2, R3 = radii**2
        n_x, n_y = norms
        RX = n_x**2
        I, J = cxy**2, cyy**2
        X = (R1 - R2 + RX) / (2*n_x)
        Y = (R1 - R3 - 2*cxy*X + I + J) / (2*cyy)
        Z = np.sqrt(R1 - X**2 - Y**2)

        return [
            centers[0] + np.dot([X, Y, Z], axes),
            centers[0] + np.dot([X, Y, -Z], axes),
        ]

    @classmethod
    def get_intersections(cls, centers, radii):
        dm = nput.distance_matrix(centers)
        nc = len(centers)
        intersects = np.full((nc, nc), True)
        intersection_disks = []
        for i, j in itertools.combinations(range(nc), 2):
            if radii[i] + radii[j] > dm[i, j]:
                intersects[i, j] = True
                intersection_disks.append(
                    cls.sphere_double_intersection_circle(
                        centers[(i, j),],
                        radii[(i, j),],
                        dist=dm[i, j]
                    )
                )
        intersection_points = []
        for i, j, k in itertools.combinations(range(nc), 3):
            if all(intersects[p] for p in itertools.combinations((i, j, k), 2)):
                w = cls._trip_w(
                    dm[j, k], dm[i, k], dm[i, j],
                    radii[i], radii[j], radii[k]
                )
                if w > 0:
                    intersection_points.append(
                        cls.sphere_triple_intersection_point(
                            centers[(i, j, k),],
                            radii[(i, j, k),],
                            dists=(dm[i, j], dm[i, k])
                        )
                    )

        return intersection_points, intersection_disks

    @classmethod
    def sphere_double_intersection_area(cls, a, r1, r2):
        t1 = r1 + r2 - a
        t2 = r1 - r2 + a
        t3 = -r1 + r2 + a
        if t1 > 0 and t2 > 0 and t3 > 0:
            return None, np.pi * (
                    2 * (r1**2 + r2**2)
                    - (r1 + r2) * a
                    - (r2 - r1)*(r2**2 - r1**2)/a
            )
        elif t1 < 0:
            return (), 0
        elif t2 < 0:
            return (0,), cls.sphere_area(r1)
        else:
            return (1,), cls.sphere_area(r2)

    @classmethod
    def _quad_w(cls, a, b, c, f, g, h):
        A = a ** 2
        B = b ** 2
        C = c ** 2
        F = f ** 2
        G = g ** 2
        H = h ** 2
        d = np.linalg.det([
            [0, 1, 1, 1, 1],
            [1, 0, H, G, F],
            [1, H, 0, A, B],
            [1, G, A, 0, C],
            [1, F, B, C, 0],
        ])
        if d < 0:
            return d
        else:
            return np.sqrt(1 / 2 * d)
    @classmethod
    def _quad_s(cls, a, b, c, f, g, h):
        return a * (
                b**2 + c**2 - a**2 + g**2 + h**2
                - 2*f**2 + (g**2 - h**2)*(b**2 - c**2)/a**2
        )
    @classmethod
    def _quad_term(cls, a, beta, gamma, W2, s1):
        t = np.arctan(W2 / s1)
        if t < 0:
            t = t + np.pi
        elif t > np.pi:
            t = t - np.pi
        return a*(beta + gamma)*(1 + ((beta - gamma)**2) /a**2) * t
    @classmethod
    def triangle_area(cls, a, b, c):
        s = (a + b + c) /2
        area = np.sqrt(s*(s-a)*(s-b)*(s-c))
        return area
    @classmethod
    def sphere_quadruple_intersection_area(cls,
                                           a, b, c, f, g, h,
                                           # 23, 12, 13, 14, 24, 34
                                           r1, r2, r3, r4,
                                           A123, A124, A134, A234,
                                           I4, I3, I2, I1
                                           ):


        W2 = 2*cls._quad_w(a, b, c, f, g, h)

        # if W2 < 0:
        ia4, ib4 = I4
        ia3, ib3 = I3
        ia2, ib2 = I2
        ia1, ib1 = I1

        test_bits = (int(ia4), int(ib4), int(ia3), int(ib3), int(ia2), int(ib2), int(ia1), int(ib1))

        # clean_tests = [
        #     ((1, 0) if s1 == 0 else (0, 1))
        #     + ((1, 0) if s2 == 0 else (0, 1))
        #     + ((1, 0) if s3 == 0 else (0, 1))
        #     + ((1, 0) if s4 == 0 else (0, 1))
        #     for s1, s2, s3, s4 in itertools.product(range(2), range(2), range(2), range(2))
        # ]
        clean_tests = {
            (1, 0, 1, 0, 1, 0, 1, 0),
            (1, 0, 1, 0, 1, 0, 0, 1),
            (1, 0, 1, 0, 0, 1, 1, 0),
            (1, 0, 1, 0, 0, 1, 0, 1),
            (1, 0, 0, 1, 1, 0, 1, 0),
            (1, 0, 0, 1, 1, 0, 0, 1),
            (1, 0, 0, 1, 0, 1, 1, 0),
            (1, 0, 0, 1, 0, 1, 0, 1),
            (0, 1, 1, 0, 1, 0, 1, 0),
            (0, 1, 1, 0, 1, 0, 0, 1),
            (0, 1, 1, 0, 0, 1, 1, 0),
            (0, 1, 1, 0, 0, 1, 0, 1),
            (0, 1, 0, 1, 1, 0, 1, 0),
            (0, 1, 0, 1, 1, 0, 0, 1),
            (0, 1, 0, 1, 0, 1, 1, 0),
            (0, 1, 0, 1, 0, 1, 0, 1)
        }

        # print(test_bits)

        if test_bits == (1, 1, 0, 0, 0, 0, 0, 0):
            # _, area = cls.sphere_triple_intersection_area(a, b, c, r1, r2, r3)
            return (0, 1, 2), None
        elif test_bits == (0, 0, 1, 1, 0, 0, 0, 0):
            # _, area = cls.sphere_triple_intersection_area(g, f, c, r1, r2, r4)
            return (0, 1, 3), None
        elif test_bits == (0, 0, 0, 0, 1, 1, 0, 0):
            # _, area = cls.sphere_triple_intersection_area(h, f, b, r1, r3, r4)
            return (0, 2, 3), None
        elif test_bits == (0, 0, 0, 0, 0, 0, 1, 1):
            # _, area = cls.sphere_triple_intersection_area(h, g, a, r2, r3, r4)
            return (1, 2, 3), None
        elif test_bits == (1, 1, 1, 1, 0, 0, 0, 0):
            # _, area = cls.sphere_double_intersection_area(c, r1, r2)
            return (0, 1), None
        elif test_bits == (1, 1, 0, 0, 1, 1, 0, 0):
            # _, area = cls.sphere_double_intersection_area(b, r1, r3)
            return (0, 2), None
        elif test_bits == (0, 0, 1, 1, 1, 1, 0, 0):
            # _, area = cls.sphere_double_intersection_area(f, r1, r4)
            return (0, 3), None
        elif test_bits == (1, 1, 0, 0, 0, 0, 1, 1):
            # _, area = cls.sphere_double_intersection_area(a, r2, r3)
            return (1, 2), None
        elif test_bits == (0, 0, 1, 1, 0, 0, 1, 1):
            # _, area = cls.sphere_double_intersection_area(g, r2, r4)
            return (1, 3), None
        elif test_bits == (0, 0, 0, 0, 1, 1, 1, 1):
            # _, area = cls.sphere_double_intersection_area(h, r3, r4)
            return (2, 3), None
        elif test_bits == (1, 1, 1, 1, 1, 1, 0, 0):
            return (0,), None#cls.sphere_area(r1)
        elif test_bits == (1, 1, 1, 1, 0, 0, 1, 1):
            return (1,), None#cls.sphere_area(r2)
        elif test_bits == (1, 1, 0, 0, 1, 1, 1, 1):
            return (2,), None#cls.sphere_area(r3)
        elif test_bits == (0, 0, 1, 1, 1, 1, 1, 1):
            return (3,), None#cls.sphere_area(r4)
        elif test_bits == (1, 1, 1, 1, 1, 1, 1, 1):
            # if W2 < 0:
            T123 = cls.triangle_area(r1, r2, r3)
            T124 = cls.triangle_area(r1, r2, r4)
            T134 = cls.triangle_area(r1, r3, r4)
            T234 = cls.triangle_area(r2, r3, r4)
            test_vec = np.array([T123, T124, T134, T234])

            if (
                    abs(np.dot([1, 1, -1, -1], test_vec)) < 1e-6
                    or abs(np.dot([1, -1, 1, -1], test_vec)) < 1e-6
                    or abs(np.dot([1, -1, -1, 1], test_vec)) < 1e-6
            ):
                A, B, C, F, G, H = np.array([a, b, c, f, g, h])**2
                # need to find triangles
                if (A - (B + C)) > 1e-6:
                    # 1 --c-- 2
                    # | b   g |
                    # 3 --h-- 4
                    # _, area = cls.sphere_double_intersection_area(a, r2, r3)
                    return (1, 2), None
                elif (B - (A + C)) > 1e-6:
                    # 2 --c-- 1
                    # | a   g |
                    # 3 --h-- 4
                    # _, area = cls.sphere_double_intersection_area(b, r1, r3)
                    return (0, 2), None
                elif (C - (A + B)) > 1e-6:
                    # 3 --b-- 1
                    # | a   f |
                    # 2 --g-- 4
                    # _, area = cls.sphere_double_intersection_area(c, r1, r2)
                    return (0, 1), None
                elif (F - (G + H)) > 1e-6:
                    # 3 --b-- 1
                    # | h   c |
                    # 4 --f-- 2
                    # f is a diagonal
                    # _, area = cls.sphere_double_intersection_area(f, r1, r4)
                    return (0, 3), None
                # elif (G - (F + H)) > 1e-6:
                #     # g is a diagonal
                #     _, area = cls.sphere_double_intersection_area(g, r2, r4)
                #     return (1, 3), area
                else:
                    raise ValueError("no set of diagonals works...")
            elif abs(np.dot([1, 1, 1, -1], test_vec)) < 1e-6:
                raise NotImplementedError(...)
            else:
                raise ValueError("no set of areas works...")
        elif test_bits not in clean_tests:
            raise ValueError(test_bits)

        if W2 <= 0:
            raise ValueError(...)

        s1 = cls._quad_s(a, b, c, f, g, h)
        s2 = cls._quad_s(b, c, a, g, h, f)
        # s22 = b * (
        #         c ** 2 + a ** 2 - b ** 2 + h ** 2 + f ** 2
        #         - 2 * g ** 2 + (h ** 2 - f ** 2) * (c ** 2 - a ** 2) / b ** 2
        # )
        # if s2 != s22:
        #     raise ValueError(...)
        s3 = cls._quad_s(c, a, b, h, f, g)
        # si = c * (
        #         a ** 2 + b ** 2 - c ** 2 + f ** 2 + g ** 2
        #         - 2 * h ** 2 + (f ** 2 - g ** 2) * (a ** 2 - b ** 2) / c ** 2
        # )
        # if s3 != si:
        #     raise ValueError(...)
        s4 = cls._quad_s(h, a, g, c, f, b)
        # si = h * (
        #         f ** 2 + b ** 2 - h ** 2 + a ** 2 + g ** 2
        #         - 2 * c ** 2 + (f ** 2 - b ** 2) * (a ** 2 - g ** 2) / h ** 2
        # )
        # if s4 != si:
        #     raise ValueError(...)
        s5 = cls._quad_s(g, h, a, b, c, f)
        # si = g * (
        #         h ** 2 + a ** 2 - g ** 2 + c ** 2 + f ** 2
        #         - 2 * b ** 2 + (c ** 2 - f ** 2) * (h ** 2 - a ** 2) / g ** 2
        # )
        # if s5 != si:
        #     raise ValueError(...)
        s6 = cls._quad_s(f, g, c, a, b, h)
        # si = f * (
        #         g ** 2 + c ** 2 - f ** 2 + b ** 2 + h ** 2
        #         - 2 * a ** 2 + (b ** 2 - h ** 2) * (g ** 2 - c ** 2) / f ** 2
        # )
        # if s6 != si:
        #     raise ValueError(...)

        A1 = cls._quad_term(a, r2, r3, W2, s1) #a*(beta + gamma)*(1 + ((beta - gamma)**2) /a**2) * t
        A2 = cls._quad_term(b, r1, r3, W2, s2)
        A3 = cls._quad_term(c, r1, r2, W2, s3)
        A4 = cls._quad_term(h, r3, r4, W2, s4)
        A5 = cls._quad_term(g, r2, r4, W2, s5)
        A6 = cls._quad_term(f, r1, r4, W2, s6)

        A = (
                -np.pi * (r1 ** 2 + r2 ** 2 + r3 ** 2 + r4 ** 2)
                + 1 / 2 * (
                        A1 + A2 + A3
                        + A4 + A5 + A6
                        + A123 + A124
                        + A134 + A234
                )
        )

        return None, A

    @classmethod
    def sphere_area(cls, radii, axis=None):
        return 4*np.pi*np.sum(radii**2, axis=axis)

    @classmethod
    def sphere_union_surface_area(cls,
                                  centers, radii,
                                  include_doubles=True,
                                  include_triples=None,
                                  include_quadruples=None,
                                  # include_quintuples=None,
                                  return_terms=False,
                                  overlap_tolerance=0):
        if include_triples is None:
            include_triples = include_doubles
        if include_quadruples is None:
            include_quadruples = include_triples is not False
        # if include_quintuples is None:
        #     include_quintuples = include_quadruples is not False

        dm = nput.distance_matrix(centers)
        terms = {
            (i,):v
            for i,v in enumerate(cls.sphere_area(np.asanyarray(radii)[:, np.newaxis], axis=-1))
        }
        nc = len(centers)
        # visible = np.full((nc), True)
        # include_pairs = np.full((nc, nc), False)
        if include_doubles:
            for i,j in itertools.combinations(range(nc), 2):
                if not (
                        (i,) in terms
                        and (j,) in terms
                ): continue
                if (radii[i] + radii[j]) * (1+overlap_tolerance) > dm[i, j]:
                    overlaps, contrib = cls.sphere_double_intersection_area(
                        dm[i, j], radii[i], radii[j]
                    )
                    if overlaps is not None:
                        if len(overlaps) > 0:
                            k = [i, j][overlaps[0]]
                            # visible[k] = False
                            terms.pop((k,), None)
                    else:
                        # include_pairs[i, j] = True
                        terms[(i, j)] = -contrib

        if include_quadruples:
            intersection_points = np.full((nc, nc, nc, 2, 3), 0.0)
        else:
            intersection_points = None
        if include_triples or include_quadruples:
            for i,j,k in itertools.combinations(range(nc), 3):
                if not all(p in terms for p in itertools.combinations((i, j, k), 1)): continue
                if all(p in terms for p in itertools.combinations((i, j, k), 2)):
                    overlaps, contrib = cls.sphere_triple_intersection_area(
                        dm[j, k], dm[i, k], dm[i, j],
                        radii[i], radii[j], radii[k]
                    )
                    if overlaps is not None:
                        overlaps = tuple([i,j,k][x] for x in overlaps)
                        if len(overlaps) == 2:
                            i, j = overlaps
                            term_keys = list(terms.keys())
                            for p in term_keys:
                                if i in p and j in p:
                                    del terms[p]
                        elif len(overlaps) == 1:
                            l, = overlaps
                            term_keys = list(terms.keys())
                            for p in term_keys:
                                if l in p:
                                    del terms[p]
                    else:
                        if include_quadruples:
                            intersection_points[i, j, k] = cls.sphere_triple_intersection_point(
                                centers[(i, j, k),],
                                radii[(i, j, k),],
                                dists=(dm[i, j], dm[i, k])
                            )
                        terms[(i, j, k)] = contrib

        intersection_tests = None
        include_quartics = None
        if include_quadruples:
            intersection_tests = np.full((nc, nc, nc, nc, 2), False)
            # include_quartics = np.full((nc, nc, nc, nc), False)
            for i,j,k,l in itertools.combinations(range(nc), 4):
                if not all(p in terms for p in itertools.combinations((i, j, k, l), 1)): continue
                if not all(p in terms for p in itertools.combinations((i, j, k, l), 2)): continue
                if all(p in terms for p in itertools.combinations((i, j, k, l), 3)):
                    l_ints = intersection_tests[i, j, k, l] = (
                            np.linalg.norm(centers[(l,)] - intersection_points[i, j, k], axis=-1)
                             < radii[l] * (1 + overlap_tolerance)
                    )
                    k_ints = intersection_tests[i, j, l, k] = (
                            np.linalg.norm(centers[(k,)] - intersection_points[i, j, l], axis=-1)
                            < radii[k] * (1+overlap_tolerance)
                    )
                    j_ints = intersection_tests[i, k, l, j] = (
                            np.linalg.norm(centers[(j,)] - intersection_points[i, k, l], axis=-1)
                            < radii[j] * (1+overlap_tolerance)
                    )
                    i_ints = intersection_tests[j, k, l, i] = (
                            np.linalg.norm(centers[(i,)] - intersection_points[j, k, l], axis=-1)
                            < radii[i] * (1 + overlap_tolerance)
                    )
                    overlaps, contrib = cls.sphere_quadruple_intersection_area(
                        dm[j, k], dm[i, k], dm[i, j],
                        dm[i, l], dm[j, l], dm[k, l],
                        radii[i], radii[j], radii[k], radii[l],
                        terms[(i, j, k)], terms[(i,j,l)],
                        terms[(i, k, l)], terms[(j,k,l)],
                        l_ints, k_ints, j_ints, i_ints
                    )
                    if overlaps is not None:
                        overlaps = tuple([i, j, k, l][x] for x in overlaps)
                        if len(overlaps) == 3:
                            x, y, z = overlaps
                            term_keys = list(terms.keys())
                            for p in term_keys:
                                if x in p and y in p and z in p:
                                    del terms[p]
                        elif len(overlaps) == 2:
                            x, y, = overlaps
                            term_keys = list(terms.keys())
                            for p in term_keys:
                                if x in p and y in p:
                                    del terms[p]
                        elif len(overlaps) == 1:
                            term_keys = list(terms.keys())
                            x, = overlaps
                            for p in term_keys:
                                if x in p:
                                    del terms[p]
                    else:
                        terms[(i, j, k, l)] = -contrib

        if return_terms:
            return terms
        else:
            if include_doubles and include_triples and include_quadruples:
                return sum(terms.values())
            else:
                return sum(
                    v
                    for k,v in terms.items()
                    if (
                        len(k) == 1
                        or (include_doubles and len(k) == 2)
                        or (include_triples and len(k) == 3)
                        or (include_quadruples and len(k) == 4)
                    )
                )

    def surface_area(self, method='union', **opts):
        if method == 'union':
            return self.sphere_union_surface_area(self.centers, self.radii, **opts)
        elif method == 'sampling':
            expansion = opts.pop('expansion', self.expansion)
            scaling = opts.pop('scaling', self.scaling)
            generator_args = {
                'samples':opts.pop('samples', self.samples),
                'generator':opts.pop('sphere_point_generator', None)
            }
            opts['tolerance'] = opts.get('tolerance', self.tolerance)
            return self.sampling_point_surface_area(
                self.centers,
                self.radii * scaling + expansion,
                generator_args=generator_args,
                **opts
            )
        elif method == 'mesh':
            return self.generate_mesh().surface_area()
        else:
            raise ValueError(f"unknown surface area method '{method}'")

    def plot(self,
             figure=None,
             *,
             points=None,
             function=None,
             sphere_color='white',
             sphere_style=None,
             point_style=None,
             point_values=None,
             distance_units='Angstroms',
             plot_intersections=False,
             **etc
             ):

        if points is None:
            points = self.sampling_points

        # TODO: move this to molecule specific class...
        conv = UnitsData.convert("BohrRadius", distance_units)
        if point_values is None and function is not None:
            point_values = function(points)

        if point_values is not None:
            etc['color'] = etc.get('color', None)

        return self.plot_sphere_points(
            points * conv,
            self.centers * conv,
            self.radii * conv,
            figure=figure,
            sphere_color=sphere_color,
            point_values=point_values,
            sphere_style=sphere_style,
            point_style=point_style,
            plot_intersections=plot_intersections,
            **etc
        )

    @classmethod
    def plot_sphere_points(cls,
                           points,
                           centers,
                           radii,
                           figure=None,
                           *,
                           color='black',
                           # transparency=.8,
                           backend='x3d',
                           return_objects=False,
                           sphere_color='white',
                           sphere_style=None,
                           point_colors=None,
                           point_values=None,
                           vertex_colormap='WarioColors',
                           rescale_color_values=True,
                           plot_intersections=False,
                           intersection_point_style=None,
                           intersection_circle_style=None,
                           **etc):
        from ... import Plots as plt

        if figure is None:
            figure = plt.Graphics3D(backend=backend)

        # func_vals = func(verts)
        if point_values is not None and rescale_color_values:
            if dev.is_list_like(rescale_color_values):
                point_values = nput.vec_rescale(point_values, rescale_color_values)
            else:
                point_values = nput.vec_rescale(point_values)

            colormap = plt.Colors.ColorPalette(vertex_colormap)
            point_colors = colormap(point_values)

        objs = []
        if sphere_style is None:
            sphere_style = etc if color is None else {}
            if sphere_color is not None:
                sphere_style['color'] = sphere_color
            for c,r in zip(centers,radii):
                objs.append(
                    plt.Sphere(c, r, **sphere_style)
                )

        if plot_intersections:
            intersection_points, intersection_disks = cls.get_intersections(
                centers,
                radii
            )

            if intersection_circle_style is None:
                intersection_circle_style = {
                    'line_color': 'black',
                    'line_thickness':.02,
                    'subdivision':'64,64'
                }
            for disk in intersection_disks:
                circle_obj = plt.Disk(
                    disk.center,
                    radius=disk.radius,
                    normal=disk.normal,
                    **intersection_circle_style
                )
                objs.append(circle_obj)

            if len(intersection_points) > 0:
                if intersection_point_style is None:
                    intersection_point_style = {
                        'color':'red',
                        'point_size':30
                    }
                # for p in np.concatenate(intersection_points, axis=0):
                #     objs.append(plt.Sphere(p, .1, color='teal'))

                point_obj = plt.Point(
                    np.concatenate(intersection_points, axis=0),
                    **intersection_point_style
                )
                objs.append(point_obj)


        if point_colors is not None or color is not None:
            point_obj = plt.Point(
                points,
                color=color,
                vertex_colors=point_colors,
                **etc)
            objs.append(point_obj)


        for o in objs:
            o.plot(figure)

        if return_objects:
            return figure, objs
        else:
            return figure

class SphereUnionSurfaceMesh:
    def __init__(self, verts, inds, surf=None, densities=None):
        self.surf = surf
        self.verts = verts
        self.inds = inds
        self.densities = densities

    def surface_area(self, return_components=False):
        dm = nput.distance_matrix(self.verts)
        a = dm[self.inds[:, 0], self.inds[:, 1]]
        b = dm[self.inds[:, 1], self.inds[:, 2]]
        c = dm[self.inds[:, 0], self.inds[:, 2]]
        s = (a + b + c) /2
        tris = np.sqrt(s*(s-a)*(s-b)*(s-c))
        if return_components:
            return tris
        else:
            return np.sum(tris)

    @classmethod
    def from_o3d(cls, mesh, densities=None, surf=None):
        return cls(
            np.array(mesh.vertices),
            np.array(mesh.triangles),
            densities=densities,
            surf=surf
        )

    def plot(self,
             figure=None,
             *,
             function=None,
             vertex_values=None,
             distance_units='Angstroms',
             **etc
             ):

        # TODO: move this to molecule specific class...
        conv = UnitsData.convert("BohrRadius", distance_units)
        if vertex_values is None and function is not None:
            vertex_values = function(self.verts)

        if vertex_values is not None:
            etc['color'] = etc.get('color', None)

        return self.plot_triangle_mesh(
            self.verts * conv,
            self.inds,
            figure=figure,
            vertex_values=vertex_values,
            **etc
        )

    @classmethod
    def plot_triangle_mesh(cls,
                           verts, indices,
                           figure=None,
                           *,
                           color='blue',
                           transparency=.8,
                           backend='x3d',
                           return_objects=False,
                           line_color='black',
                           line_transparency=.9,
                           line_style=None,
                           vertex_colors=None,
                           vertex_values=None,
                           vertex_colormap='WarioColors',
                           rescale_color_values=True,
                           **etc):
        from ... import Plots as plt

        if figure is None:
            figure = plt.Graphics3D(backend=backend)

        # func_vals = func(verts)
        if vertex_values is not None and rescale_color_values:
            if dev.is_list_like(rescale_color_values):
                vertex_values = nput.vec_rescale(vertex_values, rescale_color_values)
            else:
                vertex_values = nput.vec_rescale(vertex_values)

            colormap = plt.Colors.ColorPalette(vertex_colormap)
            vertex_colors = colormap(vertex_values)

        objs = []
        if vertex_colors is not None or color is not None:
            tri_obj = plt.Triangle(verts,
                                   indices=indices, transparency=transparency, color=color,
                                   vertex_colors=vertex_colors,
                                   **etc)
            objs.append(tri_obj)

        if line_style is None:
            line_style = etc if color is None else {}
        else:
            line_style = line_style.copy()

        if line_color is not None or len(line_style) > 0:
            line_style['color'] = line_style.get('color', line_color)
            line_style['transparency'] = line_style.get('transparency', line_transparency)
            line_obj = plt.Line(verts, indices=indices, **line_style)
            objs.append(line_obj)

        for o in objs:
            o.plot(figure)

        if return_objects:
            return figure, objs
        else:
            return figure




from .CommonData import DataHandler, DataRecord

__all__ = [ "PotentialData" ]
__reload_hook__ = [".CommonData"]

class PotentialDataHandler(DataHandler):
    def __init__(self):
        super().__init__("PotentialData", record_type=PotentialDataRecord)
    def __getitem__(self, item):
        """
        :param item:
        :type item: str
        :return:
        :rtype: PotentialDataRecord
        """
        return super().__getitem__(item)
class PotentialDataRecord(DataRecord):
    """
    Represents a simple callable wavefunction...
    """
    def __init__(self, data_handler, key, records):
        super().__init__(data_handler, key, records)
        self.parameters = dict(self['Defaults'])
    def __call__(self, *args, **kwargs):
        kwargs = dict(self.parameters, **kwargs)
        return self['Function'](*args, **kwargs)
    def __repr__(self):
        return "{}({})".format(self.key, ", ".join("{}={}".format(k,v) for k,v in self.parameters.items()))
PotentialData=PotentialDataHandler()
PotentialData.__doc__ = """An instance of PotentialDataHandler that can be used for looking up data on potentials"""
PotentialData.__name__ = "PotentialData"


"""
Defines options for different numerical things
"""

__all__ = [
    "Options"
]

import numpy as np

class OptionsContainer:
    """
    A singleton Options object that can be used to configure options for numerical stuff
    """

    NORM_ZERO_THRESH = None
    ZERO_THRESHOLD = 1.0e-8
    ZERO_PLACEHOLDER = None

    @property
    def zero_threshold(self):
        # just here so I can modify later
        return self.ZERO_THRESHOLD
    @zero_threshold.setter
    def zero_threshold(self, v):
        self.ZERO_THRESHOLD = v

    @property
    def norm_zero_threshold(self):
        if self.NORM_ZERO_THRESH is None:
            return self.zero_threshold
        else:
            return self.NORM_ZERO_THRESH
    @norm_zero_threshold.setter
    def norm_zero_threshold(self, v):
        self.NORM_ZERO_THRESH = v

    @property
    def zero_placeholder(self):
        if self.ZERO_PLACEHOLDER is None:
            return self.zero_threshold
        else:
            return self.ZERO_PLACEHOLDER
    @zero_placeholder.setter
    def zero_placeholder(self, v):
        self.ZERO_PLACEHOLDER = v

Options = OptionsContainer()
Options.__name__ = "Options"
Options.__doc__ = OptionsContainer.__doc__



"""
Exposes some helpful utilities for creating and communicating with TCP-based persistent servers
"""

__all__ = []
from .NodeCommServer import *; from .NodeCommServer import __all__ as exposed
__all__ += exposed
from .GitServer import *; from .GitServer import __all__ as exposed
__all__ += exposed
from .EvaluationServer import *; from .EvaluationServer import __all__ as exposed
__all__ += exposed